# -*- coding: utf8 -*-
"""

***********************
``environmentsetup.py``
***********************

`BSD`_ © 2018-2019 Science and Technology Facilities Council & contributors

.. _BSD: _static/LICENSE

``environmentsetup.py`` is a module provided to configure the host environment for each project, setting up
environment variables which persist during the execution of the script which defined paths and license
references for each tool using the the flow. Specific values are extracted from the corresponding
``settings.yml`` YAML file and the tool installation  structure is determined from the ``config.yml`` YAML
file, each passed to the module via :mod:`arguments` (see :ref:`arguments`).

Todo:
    * Add ``breathe`` configuration to :meth:`ProjectEnvironment.configure_documentation_tool`

"""
import helpers.customlogging as log
import helpers.version as my_version
import helpers.arguments as arguments
import helpers.funcs as funcs

import fpgavendor_iface as fpgavendor_iface

import os

logger = log.config_logger(name=__name__)
LOGLEVEL = log.LogLevelsConsts

rev = filter(str.isdigit, "$Rev$")  # Do Not Modify this Line
version = my_version.Version(1, 0, 0, svn_rev=rev, disable_svn_logging=True)
__version__ = version.get_version()
__str__ = my_version.about(__name__, __version__, version.revision)


class ProjectEnvironment(object):
    """
    An object containing attributes defining the environment variables required for processing a project.

    Specific values are extracted from the corresponding ``settings.yml`` YAML file and the tool installation
    structure is determined from the ``config.yml`` YAML file, each passed to the module via
    :mod:`arguments` (see :ref:`arguments`).

    """
    def __init__(self, **kwargs):
        self.logger = log.config_logger(name=__name__, class_name=self.__class__.__name__)
        funcs.set_kwargs(self, kwargs)
        args = arguments.Arguments()
        setattr(self, 'args', args)
        for attr in vars(args):
            self.logger.debug('Promoting Attribute: {attr}: {value}'
                              .format(attr=attr, value=getattr(args, attr)))
            if attr == 'clean':
                clean = list()
                if isinstance(getattr(args, attr), list):
                    for c in getattr(args, attr):
                        if isinstance(c, list):
                            for cc in c:
                                clean.append(cc)
                        else:
                            clean.append(c)
                self.logger.error('clean: {clean}'.format(clean=clean))
                setattr(self, attr, clean)
            else:
                setattr(self, attr, getattr(args, attr))

        log.sect_break(self.logger)

        self.set_args_from_dict(key='environment',
                                subkey='required',
                                dictionary=self.config,
                                from_envvars=True)

        self.set_args_from_dict(key='environment',
                                subkey='exclude_keywords',
                                dictionary=self.config,
                                from_envvars=False)

        self.set_args_from_dict(key='repository',
                                dictionary=self.config,
                                from_envvars=False)

        self.set_args_from_dict(key='documentation',
                                dictionary=self.config,
                                from_envvars=False)

        log.sect_break(self.logger)
        self.logger.info('Processing Settings...')
        log.sect_break(self.logger)
        self.logger.debug('Processing Category from List of Valid Categories: {self.categories}'
                          .format(self=self))
        for cat in self.categories:
            if cat in list(self.settings.keys()):
                self.logger.info('Found Valid Category in Settings: {cat}'
                                 .format(cat=cat))
                setattr(self, 'category', cat)
                setattr(self, 'layout', cat + '_layout')

        self.set_args_from_dict(key=self.category,
                                dictionary=self.settings,
                                from_envvars=False)

        self.set_args_from_dict(key='documentation',
                                dictionary=self.settings,
                                from_envvars=False)

        for field in self.required:
            self.logger.debug('Processing Mandatory Field: {field}'
                              .format(field=field))
            attr = getattr(self, field, None)
            if attr:
                self.logger.debug('  {field}: {value}'
                                  .format(field=field, value=attr))
            else:
                log.mand_missing(self, field)

        if self.language in self.supported_languages:
            self.logger.info('Documentation Language: {self.language}'.format(self=self))
        else:
            self.logger.critical('Unsupported Documentation Language: {self.language}'
                                 .format(self=self))
            self.logger.critical('\tSupported Languages: {self.supported_languages}'
                                 .format(self=self))
            self.language = self.supported_languages[0]
            self.logger.critical('\tDefaulting to: {self.language}'
                                 .format(self=self))

        self.logger.info('Successfully Parsed Settings File to Complete Project Configuration')
        log.sect_break(self.logger)
        self.configure_modelsim_environment()
        self.configure_fpga_vendor_tool()
        self.configure_documentation_tool()
        log.sect_break(self.logger)

        self.logger.info('Processed Project Environment')
        log.sect_break(self.logger)

    def __repr__(self):
        return "{self.__class__.__name__}()".format(self=self)

    def __str__(self):
        return "{} is a {self.__class__.__name__} object".format(__name__, self=self)

    def check_if_supported(self, **kwargs):
        """Checks if value in ``settings.yml`` is supported in ``config.yml``

        Checks value against a list of ``supported_values`` and exits with error if not found with
        reference to the full-path of the failing ``settings.yml`` YAML used.

        Keyword Args:
            **value (str): Default value: ``None``
            **supported_values (list of str): Default value: ``list()``
            **settings_fileref (str): Full path, including filename, of settings file used.
                Default value: ``None``

        Returns:
            None

        """
        value = funcs.get_kwarg('value', kwargs, None)
        supported_values = funcs.get_kwarg('supported_values', kwargs, list())
        settings_fileref = funcs.get_kwarg('settings_fileref', kwargs, None)

        if value in supported_values:
            self.logger.debug('Using: {value}'
                              .format(value=value))
        else:
            self.logger.error('Unsupported: {value}'
                              .format(value=value))
            self.logger.error('Supported Values: {supported}'
                              .format(supported=supported_values))
            self.logger.error('Check: {settings_file}'
                              .format(settings_file=settings_fileref))
            log.errorexit(self.logger)

    def set_attr_from_path(self, **kwargs):
        """Sets an attribute: ``name`` containing a full-path, without file name.

        Keyword Args:
            **name (str): Name of the attribute to set. Default value: ``None``
            **path (str): Full path, excluding any filenames, of reference to attribute.
                Default value: ```None``
            **create_if_missing (bool, optional): If ``True`` Creates the directory structure on the
                file-system, using ``os.makedirs()``. If the directory structure does **not** exist on the
                file-system. Default value: ``False``
            **force_attr (bool, optional): If ``True`` Creates the attribute even if the directory structure
                does **not** exist on the file-system. Default value: ``False``

        Returns:
            None

        """
        name = funcs.get_kwarg('name', kwargs, None)
        path = funcs.get_kwarg('path', kwargs, None)
        create_if_missing = funcs.get_kwarg('create_if_missing', kwargs, False)
        force_attr = funcs.get_kwarg('force_attr', kwargs, False)

        if os.path.exists(path):
            self.logger.debug('Parsed:  {path}'
                              .format(path=path))
            setattr(self, name, path)
        else:
            self.logger.critical('Path Missing on File-System: {path}'
                                 .format(path=path))
            if create_if_missing:
                self.logger.warning('Creating Path on File-System: {path}'
                                    .format(path=path))
                os.makedirs(path)

            if force_attr:
                setattr(self, name, path)

    def set_args_from_dict(self, **kwargs):
        """Sets arguments from a dictionary.

        Creates attributes from a dictionary based on ``key`` or ``key[subkey]``, from
        the item or items in list found in ``dictionary[key]`` or ``dictionary[key][subkey]``.
        If using: ``from_envvars=True`` the attribute will be sourced from the system environment
        variable.

        Keyword Args:
            **key (str): Name of the dictionary key to process. Default value: ``None``
            **subkey (str, optional): Name of the dictionary sub-key to process. Default value: ``None``
            **dictionary (dict): Dictionary to extract ``[key]`` or ``[key][subkey]`` value from.
                Default value: ``dict()``
            **from_envvars (bool, optional): If ``True`` gets the value from the corresponding system
                environment variable. Default value: ``False``

        Returns:
            None

        """
        key = funcs.get_kwarg('key', kwargs, None)
        subkey = funcs.get_kwarg('subkey', kwargs, None)
        dictionary = funcs.get_kwarg('dictionary', kwargs, dict())
        from_envvars = funcs.get_kwarg('from_envvars', kwargs, False)

        if key in list(dictionary.keys()):
            self.logger.info('Setting Up {key}...'
                             .format(key=key.capitalize()))
            if from_envvars:
                if subkey:
                    self.setattr_from_envvar(envs=dictionary[key][subkey])
                else:
                    self.setattr_from_envvar(envs=dictionary[key])
            else:
                if subkey:
                    if isinstance(dictionary[key][subkey], (dict,)):
                        for k, v in dictionary[key][subkey].iteritem():
                            self.logger.debug('Adding Attribute {self.__class__.__name__}.{k} => {v}'
                                              .format(self=self, k=k, v=v))
                            setattr(self, k, v)
                    else:
                        self.logger.debug('Adding Attribute {self.__class__.__name__}.{k} => {v}'
                                          .format(self=self, k=subkey, v=dictionary[key][subkey]))
                        setattr(self, subkey, dictionary[key][subkey])
                else:
                    if isinstance(dictionary[key], (dict,)):
                        for k, v in dictionary[key].iteritems():
                            self.logger.debug('Adding Attribute {self.__class__.__name__}.{k} => {v}'
                                              .format(self=self, k=k, v=v))
                            setattr(self, k, v)
                    else:
                        self.logger.debug('Adding Attribute {self.__class__.__name__}.{k} => {v}'
                                          .format(self=self, k=key, v=dictionary[key]))
                        setattr(self, key, dictionary[key])

    def setattr_from_envvar(self, **kwargs):
        """Sets attribute from system environment variable.

        Gets the value from corresponding `$envs`` and sets attribute matching the system environment variable
        name with the value retrieved.

        Keyword Args:
            **envs (str): Name of the system environment variable to process. Default value: ``None``

        Returns:
            None

        """
        envs = funcs.get_kwarg('envs', kwargs, None)

        if isinstance(envs, (list, tuple, set, frozenset)):
            for env in envs:
                self.set_envattr(env=env)
        else:
            self.set_envattr(env=envs)

    def set_path_environ(self, **kwargs):
        """Sets system environment ``$path`` variable.

        This performs the equivalent of running:

        .. code-block:: bash

           setenv env path

        If ``$env`` already exists it will append to the existing value:

        .. code-block:: bash

           setenv env path:${env}

        Keyword Args:
            **env (str): Name of the system environment variable name. Default value: ``None``
            **path (str): System environment variable value. Default value: ``None``
            **delimiter (str, optional): Delimiter to use between multiple environment variable values
                Default value: ``':'``
            **overwrite (bool, optional): If ``True`` overwrite existing value instead of appending.
                Default value: ``False``

        Returns:
            None

        """
        env = funcs.get_kwarg('env', kwargs, None)
        path = funcs.get_kwarg('path', kwargs, None)
        delimiter = funcs.get_kwarg('delimiter', kwargs, ':')
        overwrite = funcs.get_kwarg('overwrite', kwargs, None)

        try:
            temp_env = os.environ[env.upper()]
            self.logger.debug('Adding:  {env} {value}'
                              .format(env=env.upper(), value=path))
            if overwrite:
                temp_env = None
        except KeyError:
            temp_env = None
            self.logger.debug('Setting: {env} {value}'
                              .format(env=env.upper(), value=path))

        self.logger.debug('Old Env: {env}'
                          .format(env=temp_env))

        if temp_env:
            temp_list = temp_env.split(delimiter)
            self.logger.debug('Processing Existing Environment Variable Value:')
            for i in temp_list:
                self.logger.debug('\t{i}'
                                  .format(i=i))
            if path in temp_list:
                self.logger.warning('{path} Already Set in: ${env}'
                                    .format(path=path, env=env.upper()))
                self.logger.warning('\tMoving to First Reference')
                temp_list.remove(path)
                temp_env = '{delimiter}'.format(delimiter=delimiter).join(temp_list)
                self.logger.debug('Re-ordered: {env}'
                                  .format(env=temp_env))
            temp_env = '{path}{delimiter}{env}'.format(path=path, delimiter=delimiter, env=temp_env)
        else:
            temp_env = '{path}'.format(path=path)

        self.logger.debug('Updated: {value}'
                          .format(value=temp_env))

        os.environ[env.upper()] = temp_env
        temp_env = os.environ[env.upper()]
        self.logger.info('Updated: ${env} {value}'.format(env=env.upper(), value=temp_env))

    def set_envattr(self, **kwargs):
        """Checks to see if the attribute already exists before setting.

        Keyword Args:
            **env (str): Name of the system environment variable to check. Default value: ``None``

        Returns:
            None
        """
        env = funcs.get_kwarg('env', kwargs, None)

        if hasattr(self, env.lower()):
            self.logger.crtical('Attribute Already Set: {}'
                                .format(env.lower()))
        else:
            setattr(self, env.lower(), self.get_envvar(variable_name=env))

    def get_envvar(self, **kwargs):
        """Method to get system environment variable using ``os.environ``

        If the system environment variable does **not** exist :meth:`customlogging.errorexit` is called.

        Keyword Args:
            **variable_name (str): Name of the system environment variable to get from system.
                Default value: ``None``

        Returns:
            (str): The retrieved system environment variable.

        """
        variable_name = funcs.get_kwarg('variable_name', kwargs, None)

        try:
            resolved_name = os.environ[variable_name]
            self.logger.info('\t${:<24}{}'
                             .format(variable_name, resolved_name))
            return resolved_name
        except KeyError:
            self.logger.error('Environment Variable Not Set: {name}'
                              .format(name=variable_name.upper()))
            self.logger.error('\tFor Linux Add Environment Variable Using:')
            self.logger.error('\t\tsetenv {name} <VALUE>'
                              .format(name=variable_name.upper()))
            self.logger.error('\tWhere <VALUE> is the Variable Value')
            log.errorexit(self.logger)

    def add_environment_variable(self, **kwargs):
        """Adds a system environment variable for the running script.

        If the system environment variable already exists this will add the new value as the first value
        and append the old value using ``':'`` as a delimiter.

        Keyword Args:
            **name (str): Name of the system environment variable to create/append. Default value: ``None``
            **value (str): Value to assign to the system environment variable. Default value: ``None``
            **overwrite (bool, optional): Default value: ``False``

        Returns:
            None

        """
        name = funcs.get_kwarg('name', kwargs, None)
        value = funcs.get_kwarg('value', kwargs, None)
        overwrite = funcs.get_kwarg('overwrite', kwargs, False)

        self.logger.debug('Setting: {name} -> {value}'
                          .format(name=name.upper(), value=value))
        self.set_path_environ(env=name.upper(), path=value, overwrite=overwrite)

    def configure_modelsim_environment(self):
        """Configures and validates the system environment variables for running ``modelsim``

        Returns:
            None

        """
        self.logger.info('Setting Up Simulation Environment...')
        self.check_if_supported(value=self.sim_vendor,
                                supported_values=self.config['simulation']['supported_vendors'],
                                settings_fileref=self.settings_file)
        self.logger.info('Simulator: {vendor}'
                         .format(vendor=self.sim_vendor.capitalize()))

        self.check_if_supported(value=self.sim_version,
                                supported_values=self.config['simulation'][self.sim_vendor]['supported_tools'],
                                settings_fileref=self.settings_file)
        self.logger.info('Using Version: {version}'
                         .format(version=self.sim_version))

        self.configure_license(config_dict=self.config['simulation'][self.sim_vendor]['license'])

        try:
            self.configure_mode(mode=self.sim_mode,
                                config_dict=self.config['simulation'][self.sim_vendor]['mode'],
                                mode_type='sim')
        except AttributeError:
            setattr(self, 'sim_mode', 32)
            self.logger.info('Sim Mode Defaulting to: {mode}'
                             .format(mode=self.sim_mode))
            self.configure_mode(mode=self.sim_mode,
                                config_dict=self.config['simulation'][self.sim_vendor]['mode'],
                                mode_type='sim')

        self.configure_paths(config_dict=self.config['simulation'][self.sim_vendor]['paths'])

        setattr(self, 'sim_version', self.sim_vendor + '_' + self.sim_version.replace('/', '_'))
        self.logger.info('Renamed Simulation Version for Build Referencing: {self.sim_version}'
                         .format(self=self))

        self.logger.info('Successfully Configured Simulation Environment')
        log.sect_break(self.logger)

    def configure_fpga_vendor_tool(self):
        """Configures and validates the system environment variables for running the `FPGA` vendor tool.

        Returns:
            None

        """
        self.logger.info('Setting Up FPGA Vendor Environment...')
        self.check_if_supported(value=self.vendor,
                                supported_values=self.config['fpga']['supported_vendors'],
                                settings_fileref=self.settings_file)
        self.logger.info('FPGA Vendor: {vendor}'
                         .format(vendor=self.vendor.capitalize()))

        self.check_if_supported(value=self.tool_version,
                                supported_values=self.config['fpga'][self.vendor]['tools'],
                                settings_fileref=self.settings_file)
        self.logger.info('Using Version: {version}'
                         .format(version=self.tool_version))

        self.configure_license(config_dict=self.config['fpga'][self.vendor]['license'])

        try:
            self.configure_mode(mode=self.tool_mode,
                                config_dict=self.config['fpga'][self.vendor]['mode'],
                                mode_type='tool')
        except AttributeError:
            tool_mode = self.config['fpga'][self.vendor]['mode']['supported_modes'][0]
            self.logger.debug('Setting Tool Mode: {mode}'
                              .format(mode=tool_mode))
            setattr(self, 'tool_mode', tool_mode)
            self.logger.info('Tool Mode Defaulting to: {mode}'
                             .format(mode=self.tool_mode))
            self.configure_mode(mode=self.tool_mode,
                                config_dict=self.config['fpga'][self.vendor]['mode'],
                                mode_type='tool')

        self.configure_paths(config_dict=self.config['fpga'][self.vendor]['paths'])

        setattr(self, 'tool_version', fpgavendor_iface.resolve_tool_version(tool_version=self.tool_version,
                                                                            vendor=self.vendor))

        self.logger.info('Renamed Tool Version for Build Referencing: {self.tool_version}'
                         .format(self=self))

        self.logger.info('Successfully Configured FPGA Vendor Tool Environment')
        log.sect_break(self.logger)

    def configure_documentation_tool(self):
        """Configures and validates the system environment variables for running documentation tools.

        Supports the following automatic documentation generation tools:

        * ``doxygen``
        * ``breathe``
        * ``sphinx``

        Returns:
            None

        """
        self.logger.info('Setting Up Documentation Environment...')
        self.logger.info('Configuring Doxygen...')
        self.check_if_supported(value=self.doxy_version,
                                supported_values=self.config['documentation']['doxygen']['supported_tools'],
                                settings_fileref=self.settings_file)
        self.logger.info('\tUsing Version: {version}'
                         .format(version=self.doxy_version))
        self.configure_paths(config_dict=self.config['documentation']['doxygen']['paths'])

        self.logger.info('Configuring Dot...')
        self.check_if_supported(value=self.dot_version,
                                supported_values=self.config['documentation']['dot']['supported_tools'],
                                settings_fileref=self.settings_file)
        self.logger.info('\tUsing Version: {version}'
                         .format(version=self.dot_version))
        self.configure_paths(config_dict=self.config['documentation']['dot']['paths'])

        self.logger.info('Configuring Breathe...')
        # \TODO Configure Breathe Here...

        self.logger.info('Configuring Sphinx...')
        self.check_if_supported(value=self.sphinx_version,
                                supported_values=self.config['documentation']['sphinx']['supported_tools'],
                                settings_fileref=self.settings_file)
        self.logger.info('\tUsing Version: {version}'
                         .format(version=self.sphinx_version))
        self.configure_paths(config_dict=self.config['documentation']['sphinx']['paths'])

    def configure_license(self, **kwargs):
        """Configures license environment variables.

        Gets configuration from ``config.yml`` YAML file (see :ref:`configuration`).

        Keyword Args:
            **license_settings (dict): The license settings from configuration dictionary.
                Default value: ``None``

        Returns:
            None

        """
        license_settings = funcs.get_kwarg('config_dict', kwargs, None)

        if isinstance(license_settings, (dict,)):
            for k, v in license_settings.iteritems():
                self.logger.info('Setting License: {k}: {v}'
                                 .format(k=k, v=v))
                setattr(self, k, v)
                if isinstance(v, (list,)):
                    for i in v:
                        self.add_environment_variable(name=k, value=i)
                else:
                    self.add_environment_variable(name=k, value=v)

    def configure_mode(self, **kwargs):
        """Configure tool execution mode.

        Provides support for running the tools in 32 or 64 bit modes as well as referencing libraries not
        supported by the host operating system distribution.

        Keyword Args:
            **mode (str): Default value: ``None``
            **mode_config (dict): Dictionary defining the mode of running the tool being configured.
                Default value: ``None``
            **mode_type (str): Default value: ``sim``
        Returns:
            None

        """
        mode = funcs.get_kwarg('mode', kwargs, None)
        mode_config = funcs.get_kwarg('config_dict', kwargs, None)
        mode_type = funcs.get_kwarg('mode_type', kwargs, 'sim')

        try:
            if mode in mode_config['supported_modes']:
                self.logger.info('{mode_type} Mode: {mode}'
                                 .format(mode_type=mode_type.capitalize(),
                                         mode=mode))
            else:
                mode = mode_config['supported_modes'][0]
                self.logger.debug('Setting Mode to: {mode}'
                                  .format(mode=mode))
                setattr(self, mode_type.lower() + '_mode', mode)
                self.logger.warning('Unsupported {mode_type} Defaulting to: {mode}'
                                    .format(mode_type=mode_type.capitalize(),
                                            mode=mode))
                self.logger.warning('\tSupported {mode_type} Modes: {modes}'
                                    .format(mode_type=mode_type.capitalize(),
                                            modes=mode_config['supported_modes']))

            self.add_environment_variable(name=mode_config['env_name'], value=mode, overwrite=True)
            ld_library_path = mode_config['ld_library_path_' + str(mode)]
        except TypeError:
            ld_library_path = mode_config['ld_library_path']

        try:
            if isinstance(ld_library_path, (list,)):
                for p in ld_library_path:
                    self.add_environment_variable(name='ld_library_path', value=p, overwrite=False)
        except KeyError:
            self.logger.info('No Additional LD_LIBRARY_PATH References to Add. Skipping')

    def configure_paths(self, **kwargs):
        """Configures system environment paths required for tool execution.

        Keyword Args:
            **paths_config (dict): Dictionary defining the paths and values for the tool being configured.
                Default value: ``None``

        Returns:
            None

        """
        paths_config = funcs.get_kwarg('config_dict', kwargs, None)

        try:
            root_path = funcs.replace_tag_with_attr_value(self, paths_config['root_path'])
            root_name = paths_config['root_name']
            self.logger.debug('Root Path: {path}'
                              .format(path=root_path))
            self.logger.debug('Root Name: {name}'
                              .format(name=root_name))

            self.set_attr_from_path(name=root_name,
                                    path=root_path,
                                    create_if_missing=False)
            self.add_environment_variable(name=root_name, value=root_path, overwrite=True)

            additional_paths = paths_config.get('additional_paths', None)
            if isinstance(additional_paths, (dict,)):
                for k, v in additional_paths.iteritems():
                    self.logger.info('Setting Additional Path Environment Variables: ${k}'
                                     .format(k=k.upper()))
                    if isinstance(v, (list,)):
                        attr_value = list()
                        for i in v:
                            resolved_path = funcs.replace_tag_with_attr_value(self, i)
                            resolved_path = os.path.join(root_path, resolved_path)
                            attr_value.append(resolved_path)
                        self.set_attr_from_path(name=k,
                                                path=resolved_path,
                                                create_if_missing=False)
                        self.add_environment_variable(name=k, value=resolved_path, overwrite=False)
                        self.logger.debug('Set Attribute: {attr}: {value}'
                                          .format(attr=k, value=attr_value))
                    else:
                        resolved_path = funcs.replace_tag_with_attr_value(self, v)
                        resolved_path = os.path.join(root_path, resolved_path)
                        self.set_attr_from_path(name=k,
                                                path=resolved_path,
                                                create_if_missing=False)
                        self.add_environment_variable(name=k, value=resolved_path, overwrite=True)
                        self.logger.debug('Set Attribute: {attr}: {value}'
                                          .format(attr=k, value=resolved_path))

            add_to_path = paths_config.get('add_to_path', list())
            if isinstance(add_to_path, (list,)):
                for i in add_to_path:
                    resolved_path = funcs.replace_tag_with_attr_value(self, i)
                    resolved_path = os.path.join(root_path, resolved_path)
                    self.set_attr_from_path(name='path',
                                            path=resolved_path,
                                            create_if_missing=False)
                    self.add_environment_variable(name='path', value=resolved_path, overwrite=False)
            else:
                resolved_path = funcs.replace_tag_with_attr_value(self, add_to_path)
                resolved_path = os.path.join(root_path, resolved_path)
                self.set_attr_from_path(name='path',
                                        path=resolved_path,
                                        create_if_missing=False)
                self.add_environment_variable(name='path', value=resolved_path, overwrite=True)

            external_references = paths_config.get('external_references', None)
            if isinstance(external_references, (dict,)):
                for k, v in external_references.iteritems():
                    self.logger.info('Setting External Reference Attributes: {k}'
                                     .format(k=k.lower()))
                    if isinstance(v, (list,)):
                        attr_value = list()
                        for i in v:
                            resolved_path = funcs.replace_tag_with_attr_value(self, i)
                            resolved_path = os.path.join(root_path, resolved_path)
                            attr_value.append(resolved_path)
                        self.set_attr_from_path(name=k,
                                                path=resolved_path,
                                                create_if_missing=False,
                                                force_attr=True)
                        setattr(self, k, attr_value)
                        self.logger.debug('Set Attribute: {attr}: {value}'
                                          .format(attr=k, value=attr_value))
                    else:
                        resolved_path = funcs.replace_tag_with_attr_value(self, v)
                        resolved_path = os.path.join(root_path, resolved_path)
                        self.set_attr_from_path(name=k,
                                                path=resolved_path,
                                                create_if_missing=False,
                                                force_attr=True)
                        self.logger.debug('Set Attribute: {attr}: {value}'
                                          .format(attr=k, value=resolved_path))
        except KeyError:
            pass

        try:
            internal_references = paths_config['internal_references']
            root_path = os.path.abspath('./')
            self.logger.debug('Internal Path: {path}'
                              .format(path=root_path))
            if isinstance(internal_references, (dict,)):
                for k, v in internal_references.iteritems():
                    self.logger.info('Setting Internal Reference Attributes: {k}'
                                     .format(k=k.lower()))
                    if isinstance(v, (list,)):
                        attr_value = list()
                        for i in v:
                            resolved_path = funcs.replace_tag_with_attr_value(self, i)
                            resolved_path = os.path.join(root_path, resolved_path)
                            attr_value.append(resolved_path)
                        self.set_attr_from_path(name=k,
                                                path=resolved_path,
                                                create_if_missing=False,
                                                force_attr=True)
                        setattr(self, k, attr_value)
                        self.logger.debug('Set Attribute: {attr}: {value}'
                                          .format(attr=k, value=attr_value))
                    else:
                        resolved_path = funcs.replace_tag_with_attr_value(self, v)
                        resolved_path = os.path.abspath(os.path.join(root_path, resolved_path))

                        self.set_attr_from_path(name=k,
                                                path=resolved_path,
                                                create_if_missing=False,
                                                force_attr=True)
                        self.logger.debug('Set Attribute: {attr}: {value}'
                                          .format(attr=k, value=resolved_path))
        except KeyError:
            pass


if __name__ == '__main__':
    logger.info("environmentsetup {} {}"
                .format(__version__, version.revision))
    log.sect_break(logger)

    project = ProjectEnvironment()

else:
    logger.info(__str__)



# -*- coding: utf8 -*-
"""

*********************
``projectmanager.py``
*********************

`BSD`_ © 2018-2019 Science and Technology Facilities Council & contributors

.. _BSD: _static/LICENSE

``projectmanager.py`` is a module provided to setup, validate and execute source-code retrieval from
externally hosted repositories. Project specific values are extracted from the corresponding ``settings.yml``
YAML file and the projects structure is determined from the ``config.yml`` YAML file, each passed to the
module via :mod:`arguments` (see :ref:`arguments`).

Todo:
    * Add ``git`` support for cloning projects.
    * ``echo $PASSWORD | git svn clone $REMOTE_URL --stdlayout
      --preserve-empty-dirs --username=$REPO_USERNAME $LOCAL_PATH``
      where ``$REMOTE_URL`` doesn't include ``$sub_path`` value as this is determined using ``--stdlayout``
    * Allow projects to be hosted by a mixture of ``git`` and ``SVN``
    * Check if :meth:`ProjectDependency._layout_tag_attr_replace` can be safely replaced by
      :meth:`funcs.replace_tag_with_attr_value`

"""
import helpers.customlogging as log
import helpers.version as my_version
import environmentsetup as projectenvironment
import fpgavendor_iface as fpgavendor_iface
import helpers.funcs as funcs

import importlib
import sys
import os
import re
import posixpath
import urlparse
import getpass
import yaml

logger = log.config_logger(name=__name__)
LOGLEVEL = log.LogLevelsConsts

rev = filter(str.isdigit, "$Rev$")  # Do Not Modify this Line
version = my_version.Version(1, 2, 0, svn_rev=rev, disable_svn_logging=True)
__version__ = version.get_version()
__str__ = my_version.about(__name__, __version__, version.revision)


class RepoFlow(projectenvironment.ProjectEnvironment):
    """
    Parses the project settings file for repository_config key and processes locations found under the key.
    This provides enough information to construct the local working_copy path using ``repo_root`` inherited
    from :obj:`projectenvironment.ProjectEnvironment` and the URL of the remote repository.

    Inherits: projectenvironment.ProjectEnvironment (obj): Inherited RepoFlow object.
    See :obj:`projectenvironment.projectenvironment`.

    Username is assumed to be the same for each repository location (using self.repo_username inherited
    from projectenvironment.ProjectEnvironment)

    """
    def __init__(self, **kwargs):
        super(RepoFlow, self).__init__()
        setattr(self, 'repo_root', self.repo_root.replace('//', '/'))
        self.logger.info('REPO_ROOT: {self.repo_root}'.format(self=self))
        self.logger.info('REPO_USERNAME: {self.repo_username}'.format(self=self))

        try:
            self.logger.info('BITBUCKET_USERNAME: {self.bitbucket_username}'.format(self=self))
            bitbucket_username = self.bitbucket_username
        except AttributeError:
            bitbucket_username = ''
        log.sect_break(self.logger)

        repository_config = self.settings['repository_config']

        self.repo_types = list()
        if 'subversion' in repository_config:
            self.logger.info('Settings Contains Subversion Repository')
            self.repo_types.append('subversion')
        if 'git' in repository_config:
            self.logger.info('Settings Contains Git Repository')
            self.repo_types.append('git')
        else:
            self.logger.critical('Unknown Repository Type in Settings File: {self.settings_file}'
                                 .format(self=self))

        self.dependencies = list()
        for repo_type in self.repo_types:
            repository_config = self.settings['repository_config'][repo_type]
            self.dependencies.extend(self.process_repository_config(repository_config=repository_config,
                                                                    bitbucket_username=bitbucket_username))

        xml2vhdl_support = False
        for depend in self.dependencies:
            if depend.name == 'xml2vhdl':
                self.logger.info('Processing Environment for inclusion of "{depend.name}"...'
                                 .format(depend=depend))
                python_paths = [os.path.join(depend.scripts_dir, 'python', depend.name)]
                if os.path.exists(python_paths[0]):
                    for path in python_paths:
                        self.logger.info('Updating sys.path with: {path}'
                                         .format(path=path))
                        sys.path.append(path)
                        xml2vhdl_support = True
                        xml2vhdl_path = path
                else:
                    self.logger.critical('Can Not Update sys.path. Missing: {python_path}. '
                                         .format(python_path=python_paths[0]))

        setattr(self, 'xml2vhdl_support', xml2vhdl_support)
        setattr(self, 'xml2vhdl_path', xml2vhdl_path)

        log.sect_break(self.logger)
        self.logger.info('Processed Project Dependencies')
        log.sect_break(self.logger)

    def __repr__(self):
        return "{self.__class__.__name__}()".format(self=self)

    def __str__(self):
        return "{} is a {self.__class__.__name__} object".format(__name__, self=self)

    def process_repository_config(self, **kwargs):
        """Process repository configuration for current project

        Keyword Args:
            **repository_config (dict): Dictionary of Repository Configuration for Repository Type extracted
                from ``settings.yml`` YAML file.
            **repo_type (str): The type of repository to operate on. Valid values are: ``'subversion'`` and
                ``'git'```. Default value: ``'subversion'``
            **bitbucket_username (str, optional): The ``username`` to access ``bitbucket`` repository.
                Default value = ``''``

        Returns:
            (list of obj): A list of processed :class:`ProjectDependency` objects.

        """
        repository_config = funcs.get_kwarg('repository_config', kwargs, dict())
        repo_type = funcs.get_kwarg('repo_type', kwargs, 'subversion')
        bitbucket_username = funcs.get_kwarg('bitbucket_username', kwargs, '')
        dependencies = list()

        if isinstance(repository_config, (dict,)):
            for key, value in repository_config.items():
                self.logger.info('Processing Dependency: {name}'
                                 .format(name=key))

                if isinstance(value, (dict,)):
                    if key == self.name:
                        top_level = True
                    else:
                        top_level = False

                if value['enabled']:
                    try:
                        board_name = self.board
                    except AttributeError:
                        board_name = None

                    try:
                        board_settings = self.board_settings
                    except AttributeError:
                        board_settings = None

                    try:
                        family = self.family
                    except AttributeError:
                        family = None

                    try:
                        device = self.device
                    except AttributeError:
                        device = None

                    try:
                        vunit_default = value['vunit_default']
                    except KeyError:
                        vunit_default = False

                    try:
                        required_vhdl_libs = value['required_vhdl_libs']
                    except KeyError:
                        required_vhdl_libs = list()

                    url = value['url']
                    remote_path = self.construct_remote_path(**value)

                    try:
                        remote_rev = value['rev']
                    except KeyError:
                        remote_rev = ''

                    local_path = self.construct_local_path(root=self.repo_root,
                                                           path=value['path'])

                    category = self.get_category_from_path(supported_categories=self.categories,
                                                           category_mapping=self.category_mapping,
                                                           path=value['path'])

                    value.update(name=key)
                    if category == 'library':
                        cat_mapping = self.category_mapping
                        value.update(contrib_lib=self.get_contrib_lib_from_path(name=key,
                                                                                category_mapping=cat_mapping,
                                                                                path=value['path']))
                    else:
                        value.update(contrib_lib=None)

                    required_fields_dict = dict()

                    for attr in self.config['repository']['required']:
                        if attr in value:
                            required_fields_dict.update({attr: value[attr]})
                            self.logger.debug('\tUsing Attribute from Dependency: {attr}'
                                              .format(attr=attr))
                        else:
                            required_fields_dict.update({attr: getattr(self, attr)})
                            self.logger.debug('\tUsing Inherited Attribute: {attr}'
                                              .format(attr=attr))

                    dependencies.append(ProjectDependency(repo_root=self.repo_root,
                                                          name=key,
                                                          category=category,
                                                          url=url,
                                                          local_path=local_path,
                                                          remote_path=remote_path,
                                                          remote_rev=remote_rev,
                                                          required_vhdl_libs=required_vhdl_libs,
                                                          vunit_default=vunit_default,
                                                          top_level=top_level,
                                                          board_name=board_name,
                                                          board_settings=board_settings,
                                                          family=family,
                                                          device=device,
                                                          lib_append_categories=self.lib_append_categories,
                                                          bitbucket_username=bitbucket_username,
                                                          repo_config_dict=self.config['repository'],
                                                          tag_replacement_dict=required_fields_dict))

                    if not self.checkout_disabled:
                        self.logger.debug('Checking Out: {name}'
                                          .format(name=key))
                        try:
                            if url in _url_cache:
                                _url_cache = self._url_cache_handler(url=url,
                                                                     cache=_url_cache)
                        except NameError:
                            self.logger.info('Generating URL Cache: {url}'
                                             .format(url=url))
                            _url_cache = self._url_cache_handler(url=url,
                                                                 cache={url: False})
                        self._checkout(cache_dict=_url_cache, repo_type=repo_type,
                                       url=url,
                                       remote_path=remote_path,
                                       remote_rev=remote_rev,
                                       local_path=local_path)
                    else:
                        self.logger.info('Repository Checkout Disabled')
                        if os.path.exists(local_path):
                            self.logger.info('Local Path Found on File-System: {path}'
                                             .format(path=local_path))
                        else:
                            self.logger.critical('Local Path Missing on File-System: {path}'
                                                 .format(path=local_path))
                            self.logger.critical('\tDesign Might Not Compile...')
                            self.logger.critical('\tConfigure Project to Checkout Location')
                else:
                    self.logger.warning('Processing for "{depend}" Disabled in Settings File: {file_ref}'
                                        .format(depend=key,
                                                file_ref=self.settings_file))

                log.sect_break(self.logger)

        return dependencies

    def _url_cache_handler(self, **kwargs):
        """Caches Passwords for Accessed Controlled URLs.

        To stop the script from requesting passwords for each checkout, if the repository root URL is the
        same, this will cache the password so it only needs to be entered at the console once.

        Args:
            **url (str, optional): Default value: ``None``
            **url_cache_dict (dict, optional): Default value: ``None``

        Returns:
            None

        """
        url = funcs.get_kwarg('url', kwargs, None)
        url_cache_dict = funcs.get_kwarg('cache', kwargs, None)

        if url in url_cache_dict:
            self.logger.debug('URL Already a key in url_cache_dict: {url}'
                              .format(url=url))
        else:
            self.logger.debug('Adding URL as key in url_cache_dict: {url}'
                              .format(url=url))

        # Setup Password Prompt for Use by getpass:
        getpass_prompt = 'Password for {self.repo_username}@{url}:'.format(self=self, url=url)

        # svn logger DEBUG messages contain plain text passwords passed to svn.
        # To prevent the svn logger from writing DEBUG level messages in the log file disable all 'svn'
        # related loggers:
        for name, logr in log.logging.root.manager.loggerDict.iteritems():
            if 'svn' in name:
                logr.disabled = True
                self.logger.debug('Disabled svn logger: {name}:{logger}'
                                  .format(name=name, logger=logr))

        if not url_cache_dict[url]:
            self.logger.debug('No Password Cached for URL: {url}'
                              .format(url=url))
            if not self.dev_flag:
                url_cache_dict[url] = getpass.getpass(prompt=getpass_prompt)
            else:
                self.logger.critical('Running Script in Dev Mode: Password Caching Disabled')
        else:
            self.logger.debug('Password Already Cached for URL: {url}'
                              .format(url=url))

        return url_cache_dict

    def construct_remote_path(self, **kwargs):
        """Constructs the remote path in the externally hosted repository.

        Determines if the URL is ``bitbucket`` from ``urlparse.netloc``
        Uses ``urlparse`` to construct URL, where ``username`` is constructed manually at the front of
        ``urlparse.netloc``

        Uses ``posixpath`` to construct path portion of URL by joining ``urlparse.path``, ``path``
        and ``subpath``.

        Args:
            **url (str): The top-level root URL. See :ref:`settings_subversion`. Default value: ``None``
            **path (str): The path. See :ref:`settings_subversion`.  Default value: ``None``
            **subpath (str): The sub-path. See :ref:`settings_subversion`.  Default value: ``None``
            **bitbucket_username (str, optional): The ``username`` to access ``bitbucket`` repository.
                Default value = ``''``

        Returns:
            (str): Full path of the remote URL.

        """
        url = funcs.get_kwarg('url', kwargs, None)
        path = funcs.get_kwarg('path', kwargs, None)
        subpath = funcs.get_kwarg('subpath', kwargs, None)
        bitbucket_username = funcs.get_kwarg('bitbucket_username', kwargs, '')

        urlunsplit_param = ''
        urlunsplit_query = ''
        urlunsplit_frag = ''

        parsed_url = urlparse.urlsplit(url)

        if 'bitbucket' in parsed_url.netloc:
            self.logger.info('Processing a bitbucket URL: {url}'.format(url=url))

            if bitbucket_username != '':
                username_netloc = bitbucket_username + '@' + parsed_url.netloc
            else:
                username_netloc = parsed_url.netloc
            remote_path = urlparse.urlunparse((parsed_url.scheme,
                                               username_netloc,
                                               parsed_url.path,
                                               urlunsplit_param,
                                               urlunsplit_query,
                                               urlunsplit_frag))
        else:
            self.logger.info('Processing: {url}'.format(url=url))

            remote_path = urlparse.urlunparse((parsed_url.scheme,
                                               parsed_url.netloc,
                                               posixpath.join(parsed_url.path, path, subpath),
                                               urlunsplit_param,
                                               urlunsplit_query,
                                               urlunsplit_frag))

        self.logger.info('Constructed Full Checkout URL: {remote_path}'
                         .format(remote_path=remote_path))

        return remote_path

    def construct_local_path(self, **kwargs):
        """Constructs the local path for the local working-copy.

        Constructed by joining ``root`` and  ``path``.

        Args:
            **root (str): The top-level root. Usually derived from the ``$REPO_ROOT`` environment variable.
                Default value: ``None``
            **path (str): The path, derived from :ref:`settings_subversion`, but with the ``subpath:``
                removed. This allows the switching between ``trunk``, ``branches`` and ``tags`` to be
                managed by the repository client. Default value: ``None``

        Returns:
            (str): Full path of the local working-copy path.

        """
        root = funcs.get_kwarg('root', kwargs, None)
        path = funcs.get_kwarg('path', kwargs, None)

        local_path = os.path.join(root, path)
        self.logger.info('Constructed Full Working Copy Path:  {local_path}'
                         .format(local_path=local_path))

        return local_path

    def _checkout(self, cache_dict, repo_type='subversion', **kwargs):
        """Validates checkout parameters and checks out the source-code from remote location

        Args:
            cache_dict (dict): Default value: ``None``
            repo_type (str, optional): Operating on either a ``'subversion'`` or ``'git'`` repository.
                Default value: ``'subversion'``.

        Keyword Args:
            **url (str): Default value: ``None``.
            **remote_path (str): Default value: ``None``.
            **remote_rev (str): Default value: ``None``.
            **local_path (str): Default value: ``None``.

        Returns:
            (bool): ``False`` if trying to operate on a ``git`` repository.

        """
        url = funcs.get_kwarg('url', kwargs, None)
        remote_path = funcs.get_kwarg('remote_path', kwargs, None)
        remote_rev = funcs.get_kwarg('remote_rev', kwargs, None)
        local_path = funcs.get_kwarg('local_path', kwargs, None)

        wc_exists = False

        if os.path.exists(local_path):
            self.logger.warning('Local Path Already Exists for Checkout Operation: {path}'
                                .format(path=local_path))

            local_repo_obj = self._set_working_copy(local_path=local_path,
                                                    repo_type=repo_type)

            if local_repo_obj and self._check_working_copy(remote_type=repo_type,
                                                           local_repo_obj=local_repo_obj,
                                                           remote_path=remote_path):
                self.logger.warning('Working Copy Location is Already Checked Out. '
                                    'May Require Updating: {path}'
                                    .format(path=local_path))
                wc_exists = True
            else:
                return False

        if not wc_exists:
            remote_repo_obj = self._set_remote(cache_dict, repo_type, **kwargs)
            if remote_repo_obj:
                self.logger.info('Performing Check Out:')
                self.logger.info('\t{remote_path}@{remote_rev} -> {local_path}'
                                 .format(remote_path=remote_path,
                                         remote_rev=remote_rev,
                                         local_path=local_path))
                if repo_type == 'subversion':
                    remote_repo_obj.checkout(local_path, remote_rev)
                elif repo_type == 'git':
                    # \TODO Perform git Clone Here and return True
                    return False
            else:
                self.logger.critical('Problem Performing {repo_type} Checkout:'
                                     .format(repo_type=repo_type.capitalize()))
                self.logger.critical('\t{remote_path}@{remote_rev} -> {local_path}'
                                     .format(remote_path=remote_path,
                                             remote_rev=remote_rev,
                                             local_path=local_path))
                self.logger.critical('Skipping...')
                return False

    def _set_working_copy(self, repo_type='subversion', **kwargs):
        """Connects to an existing local working_copy.

        Args:
            repo_type (str, optional): Operating on either a ``'subversion'`` or ``'git'`` repository.
                Default value: ``'subversion'``.

        Keyword Args:
            **local_path (str): Default value: ``None``.

        Returns:
            (bool): ``False`` if trying to operate on a ``git`` repository.

        """
        local_path = funcs.get_kwarg('local_path', kwargs, None)

        if repo_type == 'subversion':
            import svn.local
            repo_obj = svn.local.LocalClient(local_path)

        elif repo_type == 'git':
            from git import Repo
            repo_obj = False

        return repo_obj

    def _check_working_copy(self, repo_type='subversion', **kwargs):
        """Checks existing local working_copy against expected working-working parameters.

        If a working-copy already exists in the local file-system, check to see if it matches the expected
        remote working-copy.

        Args:
            repo_type (str, optional): Operating on either a ``'subversion'`` or ``'git'`` repository.
                Default value: ``'subversion'``.

        Keyword Args:
            **local_path (str): Default value: ``None``.

        Returns:
            (bool): ``False`` if trying to operate on a ``git`` repository.

        """
        repo_obj = funcs.get_kwarg('local_repo_obj', kwargs, None)
        remote_path = funcs.get_kwarg('remote_path', kwargs, None)

        if repo_type == 'subversion':
            # Fetch the Associated URL from the Working Copy
            try:
                repo_info = repo_obj.info()
                wc_url = repo_info['url']
            except Exception as e:
                self.logger.critical('Failed to Connect with Working Copy')
                self.logger.critical('\t{msg}'
                                     .format(msg=e))
                self.logger.critical('\tSkipping...')
                return False
        elif repo_type == 'git':
            # \TODO Fetch the Associated URL from the Working Copy when using git and return True
            return False

        if wc_url == remote_path:
            self.logger.info('URL Match: Valid Working Copy:')
            self.logger.info('\t{url}'
                             .format(url=remote_path))
            return True
        else:
            self.logger.critical('URL Mismatched: Invalid Working Copy:')
            self.logger.critical('\tExpected: {url}'
                                 .format(url=remote_path))
            self.logger.critical('\tGot:      {url}'
                                 .format(url=wc_url))
            return False

    def _set_remote(self, cache_dict, repo_type='subversion', **kwargs):
        """ Sets the remote repository for performing Checkout/Clone operations.

        Args:
            cache_dict (dict): Default value: ``None``
            repo_type (str, optional): Operating on either a ``'subversion'`` or ``'git'`` repository.
                Default value: ``'subversion'``.

        Keyword Args:
            **url (str): Default value: ``None``.
            **remote_path (str): Default value: ``None``.

        Returns:
            (bool): ``False`` if trying to operate on a ``git`` repository.

        """

        url = funcs.get_kwarg('url', kwargs, None)
        remote_path = funcs.get_kwarg('remote_path', kwargs, None)

        if repo_type == 'subversion':
            import svn.remote
            remote_repo_obj = svn.remote.RemoteClient(remote_path,
                                                      username=self.repo_username,
                                                      password=cache_dict[url])
            try:
                repo_info = remote_repo_obj.info()
                self.logger.info('Successfully Connected to Remote Location: {username}@{url}'
                                 .format(username=self.repo_username, url=remote_path))
                return remote_repo_obj
            except Exception as e:
                self.logger.critical('Failed to Connect to Remote Location: {username}@{url}'
                                     .format(username=self.repo_username, url=remote_path))
                self.logger.critical('\t{msg}'
                                     .format(msg=e))
                self.logger.critical('Skipping...')
                return False

        elif repo_type == 'git':
            # \TODO Add Remote Support for git and return equiv. git remote_obj
            return False

    def get_category_from_path(self, **kwargs):
        """Gets ``category:`` from ``path``.

        The first directory in the path string represents the category.

        Keyword Args:
            **path (str): Default value: ``''``
            **category_mapping (dict, optional): Category mapping (see :ref:`config_categories`).
                Default value: ``dict()``.
            **supported_categories (list of str, optional): List of supported categories from ``config.yml``
                configuration file (see :ref:`config_categories`). Default value: ``list()``.

        Returns:
            (str): the ``category`` extracted from ``path``.

        """
        path = funcs.get_kwarg('path', kwargs, '')
        category_mapping = funcs.get_kwarg('category_mapping', kwargs, dict())
        supported_categories = funcs.get_kwarg('supported_categories', kwargs, list())

        cat = path.split('/')[0]

        for k, v in category_mapping.iteritems():
            if v == cat:
                cat = k

        if cat in supported_categories:
            self.logger.info('Category: {cat}'
                             .format(cat=cat))
            return cat
        else:
            self.logger.error('Could Not Resolve Category from Path')
            self.logger.error('\tPath: {path}'
                              .format(path=path))
            self.logger.error('\tCategory: {cat}'
                              .format(cat=cat))
            self.logger.error('\tSupported Categories: {cats}'
                              .format(cats=supported_categories))
            log.errorexit(self.logger)

    def get_contrib_lib_from_path(self, **kwargs):
        """Gets ``contrib_lib:`` from ``path``.

        Assumes (i know!) that ``contrib_lib`` exists between the ``category`` and ``name``.
        For a valid ``contrib_lib`` to exist the number of elements in ``path`` should be = 3.

        Keyword Args:
            **name (str): The project name being processed. Default value: ``''``
            **path (str): The path of the project on the file-system. Default value: ``''``
            **category_mapping (dict, optional): Category mapping (see :ref:`config_categories`).
                Default value: ``dict()``.

        Returns:
            (str): the ``contrib_lib`` extracted from ``path``.

        """

        name = funcs.get_kwarg('name', kwargs, '')
        path = funcs.get_kwarg('path', kwargs, '')
        category_mapping = funcs.get_kwarg('category_mapping', kwargs, dict())

        path_elements = path.split('/')
        self.logger.debug('Nof Path Elements: {len}'
                          .format(len=len(path_elements)))

        cat = path_elements[0]

        if len(path_elements) == 3:
            for k, v in category_mapping.iteritems():
                if v == cat:
                    cat = k

            contrib_lib = path_elements[1]
            path_name = path_elements[2]

            self.logger.debug('Category : {cat}'
                              .format(cat=cat))
            self.logger.debug('Contrib  : {contrib_lib}'
                              .format(contrib_lib=contrib_lib))
            self.logger.debug('Path Name: {path_name}'
                              .format(path_name=path_name))

            if path_elements[2] == name:
                contrib_lib = path_elements[1]
                self.logger.info('Contributor: {contrib_lib}'
                                 .format(contrib_lib=contrib_lib))
                return contrib_lib

        # If it gets this far raise a warning because no valid contrib_lib was found
        self.logger.warning('No Contributor for: {name}'
                            .format(name=name))
        return None


class ProjectDependency(object):
    """Processes project dependencies.

    Takes a dictionary and converts keys to attributes before parsing configuration to construct all
    attributes required to describe a Dependency Object

    Args:
        repo_root (str): The resolved ``$REPO_ROOT`` system environment variable. Required by:
            :mod:`projectmanager.ProjectDependency.layout_to_dir` and
            :mod:`projectmanager.ProjectDependency.prune_path_list`
        name (str): Name of dependency
        category (str): The dependency's category
        url (str): Root URL to remote repository location for the dependency
        local_path (str): Local path of the *working copy* for the dependency
        remote_path (str): Remote path, from the ``url`` for the dependency
        remote_rev (str): Specific repository revision for the dependency
        required_vhdl_libs (list of str, optional): List of required HDL libraries required by the dependency.
            Default value: ``list()``
        vunit_default (bool, optional): VUnit dependency flag. Default value: ``False``
        top_level (bool, optional): Project Top-level dependency flag. Default value: ``False``

        board (str, optional): Target board for dependency. Default value: ``None``
        board_settings (str, optional): Settings file name for target board for dependency.
            Default value: ``None``
        family (str, optional): Target FPGA Family for HDL dependency. Default value: ``None``
        device (str, optional): Target FPGA Device for HDL dependency. Default value: ``None``

    Keyword Args:
        **repo_config_dict (dict): The layout dictionary which defines the projects layout within the
            repository and working copy on the local file-system.
            Required by :mod:`projectmanager.ProjectDependency.layout_to_dir`
            Default value: ``dict()``
        **lib_append_categories (list of str): Default value: ``list()``
        **bitbucket_username (str, optional): Default value: ``None``

        **tag_replacement_dict (dict): Dictionary of required tag replacement fields required by
            :mod:`projectmanager.ProjectDependency.layout_tag_attr_replace`. This is variable based on
            ``category``
            Default value: ``dict()``
        **retag_layout (bool, optional). Retags the layout list (for 'vendor_ip'. Default value: ``True``

    """
    def __init__(self, repo_root, name, category, url, local_path, remote_path, remote_rev,
                 required_vhdl_libs=list(), vunit_default=False, top_level=False,
                 board_name=None, board_settings=None, family=None, device=None, **kwargs):

        self.logger = log.config_logger(name=__name__, class_name=self.__class__.__name__)
        setattr(self, 'category', category)
        setattr(self, 'top_level', top_level)
        setattr(self, 'required_vhdl_libs', required_vhdl_libs)
        setattr(self, 'vunit_default', vunit_default)
        setattr(self, 'url', url)
        setattr(self, 'local_path', local_path)
        setattr(self, 'remote_path', remote_path)
        setattr(self, 'remote_rev', remote_rev)

        setattr(self, 'family', family)
        setattr(self, 'device', device)

        lib_append_categories = funcs.get_kwarg('lib_append_categories', kwargs, list())
        retag_layout = funcs.get_kwarg('retag_layout', kwargs, True)

        if vunit_default:
            lib = self.get_lib_from_name(name=name,
                                         category=category,
                                         lib_append_categories=lib_append_categories)
        else:
            lib = None

        setattr(self, 'lib', lib)
        setattr(self, 'bitbucket_username', funcs.get_kwarg('bitbucket_username', kwargs, None))

        repo_config_dict = funcs.get_kwarg('repo_config_dict', kwargs, dict())
        self.logger.debug('Tag Replacements:')
        for k, v in funcs.get_kwarg('tag_replacement_dict', kwargs, dict()).items():
            self.logger.debug('\t{k}:{v}'.format(k=k, v=v))
            setattr(self, k, v)

        layout_str = self.category + '_layout'
        self.logger.debug('Using Layout: {layout}'
                          .format(layout=layout_str))

        setattr(self, '_temp_layout_list', list())
        _layout_dict = repo_config_dict[layout_str]
        self._top_dirs = list()
        self.layout_list = list()

        top_refs = self.set_top_refs(category, name, self.tool_version)

        if category == 'boards':
            setattr(self, 'is_board', True)
            if board_name:
                self.logger.info('Dependency is a Board: Using Board Name as Name: {board_name}'
                                 .format(board_name=board_name))
                setattr(self, 'name', board_name)
            else:
                self.logger.critical('Dependency is a Board: But No Board Name Supplied, Using: {name}'
                                     .format(name=name))
                setattr(self, 'name', name)
            if board_settings:
                if board_settings.endswith('.yml'):
                    setattr(self, 'board_settings', board_settings)
                else:
                    setattr(self, 'board_settings', board_settings + '.yml')
            else:
                default_settings_file = 'default_settings.yml'
                self.logger.warning('No Board Settings File Name Supplied. Using: {default_settings_file}'
                                    .format(default_settings_file=default_settings_file))
                setattr(self, 'board_settings', default_settings_file)

        else:
            setattr(self, 'name', name)
            setattr(self, 'is_board', False)
            setattr(self, 'board_settings', None)

        self.layout_to_dir(layout_dict=_layout_dict,
                           repo_root=repo_root,
                           top_refs=top_refs,
                           replace_tags=True)

        for kw in self._top_dirs:
            attr_name = self.top_dir_lookup(dir_name=kw)

            path_list = self.prune_path_list(path_list=self.layout_list,
                                             root_path=repo_root,
                                             keyword=kw)

            if path_list:
                if len(path_list) == 1:
                    self.logger.debug('Created Attribute: {attr} {value}'
                                      .format(attr=attr_name, value=path_list[0]))
                    setattr(self, attr_name, path_list[0])
                else:
                    self.logger.warning('More than One Path Found when Generating {attr}_dir: {path_list}'
                                        .format(attr=kw, path_list=path_list))
                    self.logger.warning('Created Attribute from First Value: {attr} {value}'
                                        .format(attr=attr_name, value=path_list[0]))
                    setattr(self, attr_name, path_list[0])
            else:
                self.logger.debug('Not Creating _dir Attribute...')

        if category == 'vendor_ip' and retag_layout:
            setattr(self, 'layout_dict', _layout_dict)

            retagged_layout_list = self.retag_path(term='vendor_ip',
                                                   tag='<NAME>',
                                                   paths=self.layout_list,
                                                   repo_root=repo_root)

            retagged_layout_list = self.retag_path(term=self.vendor,
                                                   tag='<VENDOR>',
                                                   paths=retagged_layout_list,
                                                   repo_root=repo_root)

            if self.tool_version:
                retagged_layout_list = self.retag_path(term=self.tool_version,
                                                       tag='<TOOL_VERSION>',
                                                       paths=retagged_layout_list,
                                                       repo_root=repo_root)

            if family:
                retagged_layout_list = self.retag_path(term=family,
                                                       tag='<FAMILY>',
                                                       paths=retagged_layout_list,
                                                       repo_root=repo_root)

            if device:
                retagged_layout_list = self.retag_path(term=device,
                                                       tag='<DEVICE>',
                                                       paths=retagged_layout_list,
                                                       repo_root=repo_root)

            setattr(self, 'layout_list', retagged_layout_list)

            for p in self.layout_list:
                self.logger.debug('Rebuilt Vendor IP Layout: {p}'.format(p=p))

        elif category == 'vendor_ip':
            pass

        else:
            pass

    def retag_path(self, **kwargs):
        """Re-tags path(s).

        Keyword Args:
            **term (str): Term to retag.
            **tag (str): Tag value to replace term with.
            **paths (str or list of str): absolute path(s) to retag. This assumes the absolute path is in the
                form: ``$REPO_ROOT/<TERM>/xx/xx/xx/<TERM>`` where the second term will be replaced with
                ``tag``.
            **repo_root (str): The actual path of the system environment variable ``$REPO_ROOT``

        Returns:
            list of str: List of retagged absolute paths. Empty ``list()`` if no repo_root supplied.
        """
        term = funcs.get_kwarg('term', kwargs, None)
        tag = funcs.get_kwarg('tag', kwargs, '<NAME>')
        paths = funcs.get_kwarg('paths', kwargs, list())
        repo_root = funcs.get_kwarg('repo_root', kwargs, None)

        if isinstance(paths, (str, )) and paths != "":
            paths = [paths]

        if repo_root:
            term_root = os.path.join(repo_root, term)
        else:
            return list()

        rebuild_layout_list = list()
        for p in paths:
            try:
                split_path = p.split(term_root)[1][1:]
            except IndexError:
                term_root = repo_root
                split_path = p.split(term_root)[1][1:]

            renamed_path = split_path.replace(term, tag)
            rebuilt_path = os.path.join(term_root, renamed_path)
            rebuild_layout_list.append(rebuilt_path)
            self.logger.debug('Rebuilt Layout: {p}'.format(p=rebuilt_path))

        return rebuild_layout_list

    def get_lib_from_name(self, **kwargs):
        """Determines the ``VHDL`` library name from project name.

        Determines the ``VHDL`` library name from the project name, ignoring if suffix already exists.
        the suffix is only appended if the ``category`` exists in the ``lib_append_categories`` list

        Keyword Args:
            **name (str): The project name being processed. Default value: ``''``
            **category (str): The ``category`` of the project being processed. Default value: ``''``
            **lib_append_categories (list of str, optional): List of categories to append ``lib_suffix``
                from ``config.yml`` configuration file (see :ref:`config_categories`). If the ``category``
                is **not** in this list the suffix will **not** be appended. Default value: ``list()``.
            **lib_suffix (str, optional): The library suffix to append. Default value: ``'_lib'``

        Returns:
            (str): The name of the ``VHDL`` library

        """
        name = funcs.get_kwarg('name', kwargs, '')
        category = funcs.get_kwarg('category', kwargs, '')
        lib_append_categories = funcs.get_kwarg('lib_append_categories', kwargs, list())
        lib_suffix = funcs.get_kwarg('lib_suffix', kwargs, '_lib')
        suffix_len = len(lib_suffix)
        suffix_idx = -suffix_len

        self.logger.debug('Determining Library from Name: {name}'
                          .format(name=name))
        self.logger.debug('Library Suffix: {suffix} [{idx}]'
                          .format(suffix=lib_suffix, idx=suffix_idx))

        if category in lib_append_categories:
            if len(name) > suffix_len:
                if name[suffix_idx:] != lib_suffix:
                    name = name + lib_suffix
            else:
                name = name + lib_suffix

        self.logger.info('Library Name: {name}'
                         .format(name=name))

        return name

    @staticmethod
    def top_dir_lookup(**kwargs):
        """Maps supported top-level directory names to known values for specified usage throughout

        Keyword Args:
            **dir_name (str): Name of directory name to attempt to map.

        Returns:
            str: Mapped name if matched. ``dir_name`` if *not* matched.

        """
        dir_name = funcs.get_kwarg('dir_name', kwargs, '')

        if dir_name in ['constraints', 'constr', 'constrs']:
            name = 'constraints'

        elif dir_name in ['docs', 'doc', 'documents', 'documentation']:
            name = 'docs'

        elif dir_name in ['netlists', 'net', 'nets']:
            name = 'netlists'

        elif dir_name in ['scripts', 'script']:
            name = 'scripts'

        elif dir_name in ['settings', 'setting']:
            name = 'settings'

        elif dir_name in ['simulation', 'sim', 'sims']:
            name = 'simulation'

        elif dir_name in ['src', 'srcs', 'sources']:
            name = 'src'

        else:
            name = dir_name

        return '{attr}_dir'.format(attr=name)

    @staticmethod
    def set_top_refs(category, name, tool_version):
        """

        Args:
            category (str): category to determine the top-level reference to return.
            name (str): returned if ``category`` is *not* ``boards`` or ``vendor_ip``
            tool_version (str): returned if ``category`` *is* ``boards`` or ``vendor_ip``

        Returns:
            list of str

        """
        if category == 'boards':
            return [tool_version]
        elif category == 'vendor_ip':
            return [tool_version]
        else:
            return [name]

    def layout_to_dir(self, **kwargs):
        """Takes a single top-level and creates the directory and file structure as a list.

        This layout structure is expected to be defined in the layout_dict keyword passed to the module.

        Uses top_refs list to look for a reference to indicate the top reference string, which is
        used to populate self._top_dirs, this can then be used to create attributes for top_level reference
        directories.

        Keyword Args:
            **layout_dict (dict): Default value: ``dict()``
            **repo_root (str): Default value: ``None``
            **top_refs (list of str): Default value: ``list()``
            **replace_tags (bool): Default value: ``True``

        Returns:
            None

        """
        layout_dict = funcs.get_kwarg('layout_dict', kwargs, dict())
        repo_root = funcs.get_kwarg('repo_root', kwargs, None)
        top_refs = funcs.get_kwarg('top_refs', kwargs, list())
        replace_tags = funcs.get_kwarg('replace_tags', kwargs, True)

        self.logger.debug('layout_to_dir()replace_tags: {replace_tags}'
                          .format(replace_tags=replace_tags))
        self.logger.debug('layout_to_dir():layout_dict: {layout_dict}'
                          .format(layout_dict=layout_dict))
        self.logger.debug('layout_to_dir():repo_root:   {repo_root}'
                          .format(repo_root=repo_root))

        if isinstance(layout_dict, (dict,)):
            for key, value in layout_dict.items():
                if replace_tags:
                    key = self.layout_tag_attr_replace(tag=key)
                    if key in top_refs:
                        if isinstance(value, (list, dict,)):
                            for i in value:
                                self._top_dirs.append(i.keys()[0])
                self.logger.debug('layout_to_dir():key:{}'.format(key).lower())
                self.logger.debug('1:layout_to_dir({}, {})'.format(value, os.path.join(repo_root, key)))
                if value:
                    self.layout_to_dir(layout_dict=value,
                                       repo_root=os.path.join(repo_root, key),
                                       top_refs=top_refs,
                                       replace_tags=replace_tags)
                else:
                    self.logger.debug(os.path.join(repo_root, key))
                    self.layout_list.append(os.path.join(repo_root, key))

        elif isinstance(layout_dict, (list,)):
            for item in layout_dict:
                if isinstance(item, (dict,)):
                    self.logger.debug('2:layout_to_dir({}, {})'.format(item, repo_root))
                    self.layout_to_dir(layout_dict=item,
                                       repo_root=repo_root,
                                       top_refs=top_refs,
                                       replace_tags=replace_tags)
                else:
                    if replace_tags:
                        item = self.layout_tag_attr_replace(tag=item)
                    self.logger.debug('os.repo_root.join({}, {})'.format(repo_root, item))
                    self.logger.debug(os.path.join(repo_root, item))
                    self.layout_list.append(os.path.join(repo_root, item))

    def layout_tag_attr_replace(self, **kwargs):
        """Replaces ``<TAG>`` with corresponding attribute

        Looks for ``<TAG>`` (enclosed in ``<>``) in tag string passed to the function, and if found
        returns a string with the tag replaced by the value in the ``self.<TAG>`` attribute.

        If the attribute exists but is not populated with a value an "empty"  string ``''`` is returned.
        This allows all <TAG> options in the layout configuration to be optional.
        Although ``self.name`` should always have a value as it is also used to determine filenames
        in the directory structure.

        Keyword Args:
            **tag (str): string to check for sub-strings enclosed in ``<>``.

        Returns:
            (str): An empty string if corresponding attribute for the ``<TAG>`` is **not** found; or the
            corresponding attribute if found; or the original ``<TAG>`` if anything else.

        """
        tag = funcs.get_kwarg('tag', kwargs, None)

        if '<' in tag and '>' in tag:
            attribute_tag = funcs.strip_tag(tag)
            self.logger.debug('Processing {tag}: {attribute_tag}'
                              .format(tag=tag, attribute_tag=attribute_tag))
        else:
            self.logger.debug('Not a Valid Tag: {tag}'
                              .format(tag=tag))
            return tag

        if hasattr(self, attribute_tag):
            if not getattr(self, attribute_tag):
                self.logger.critical('Attribute Not Found: {}'
                                     .format(attribute_tag))
                return ''
            else:
                self.logger.debug('Valid Tag {tag}: {attribute_tag}'
                                  .format(tag=tag, attribute_tag=attribute_tag))

                resolved_attr = getattr(self, attribute_tag)
                self.logger.debug('Resolved Attribute from Tag: {}'
                                  .format(resolved_attr))
                return re.sub('(<[^>]+>)', resolved_attr, tag)
        else:
            self.logger.debug('Attribute Not Found: {}'
                              .format(attribute_tag))
            return tag

    def prune_path(self, **kwargs):
        """Prunes path down towards ``root_path``.

        Takes a path and prunes that path until prune_to is found, if root_path is reached it means that
        prune_to was not found in the path.

        Returns path up to and including prune_to, or False if not found.

        Keyword Args:
            **path (str): Path to prune. Default value: ``None``
            **prune_to (str): Prune down to this path. Default value: ``None``
            **root_path (str): If ``root_path`` is reached the prune fails. Default value: ``None``

        Returns:
            (str): The found pruned path. The `bool:` ``False`` is returned if ``root_path`` is reached
            without finding the prune criteria.

        """
        path = funcs.get_kwarg('path', kwargs, None)
        prune_to = funcs.get_kwarg('prune_to', kwargs, None)
        root_path = funcs.get_kwarg('root_path', kwargs, None)

        while True:
            if os.path.split(path)[1] == prune_to:
                self.logger.debug('Pruned Path: {}'
                                  .format(path))
                return path
            elif os.path.split(path)[0] == root_path:
                self.logger.critical('{prune} Not found in: {path}'
                                     .format(prune=prune_to, path=path))
                return False
            else:
                path = os.path.split(path)[0]

    def prune_path_list(self, **kwargs):
        """Prunes list of paths based on ``keyword`` matching.

        Keyword Args:
            **path_list (list of str): List of paths to prune. Default value: ``None``
            **keyword (str): Keyword which **must** exist in ``path_list`` item for prune.
                Default value: ``''``
            **root_path (str): If ``root_path`` is reached the prune fails. Default value: ``None``

        Returns:
            (list of str): A list of pruned paths.

        """
        path_list = funcs.get_kwarg('path_list', kwargs, None)
        keyword = funcs.get_kwarg('keyword', kwargs, '')
        root_path = funcs.get_kwarg('root_path', kwargs, None)

        keyword_matches = [kw for kw in path_list if keyword in kw]

        pruned_paths = []
        for path in keyword_matches:
            pruned_path = self.prune_path(path=path, prune_to=keyword, root_path=root_path)
            if pruned_path not in pruned_paths:
                self.logger.debug('Adding Path to Pruned Paths: {}'
                                  .format(pruned_path))
                pruned_paths.append(pruned_path)

        return pruned_paths

    def set_as_board(self, **kwargs):

        top_level_dependency = funcs.get_kwarg('top_level_dependency', kwargs, None)
        board_priority_list = funcs.get_kwarg('board_priority_list', kwargs,
                                              ['vendor', 'tool_version', 'family', 'device'])

        board_settings_file = False
        self.logger.info('.' * 80)
        self.logger.info('Additional Board Processing...')
        if self.settings_dir:
            self.logger.debug('Board Settings Path: {settings_dir}'
                              .format(settings_dir=self.settings_dir))

            for root, subpath, files in os.walk(self.settings_dir):
                for f in files:
                    try:
                        if self.board_settings in f:
                            board_settings_file = os.path.join(root, f)
                            self.logger.info('Found Settings File for Board: {f}'
                                             .format(f=board_settings_file))
                            break
                    except AttributeError:
                        if 'default_settings' in os.path.splitext(f):
                            board_settings_file = os.path.join(root, f)
                            self.logger.info('Found Default Settings File for Board: {f}'
                                             .format(f=board_settings_file))
                            break

            if board_settings_file:
                with open(board_settings_file, 'r') as ymlfile:
                    setattr(self, 'board_settings', yaml.load(ymlfile))
                    self.logger.info('Loaded Board Settings from YAML File: {f}'
                                     .format(f=board_settings_file))

                    try:
                        board_tool_version = fpgavendor_iface.resolve_tool_version(tool_version=self.board_settings['tool_version'],
                                                                                   vendor=self.board_settings['vendor'])
                        self.board_settings['tool_version'] = board_tool_version
                    except KeyError:
                        pass

            self.logger.info('Using Settings from Board...')

            for key in board_priority_list:
                if key in self.board_settings:
                    if self.board_settings[key] == getattr(top_level_dependency, key, False):
                        self.logger.debug('[{key}] Project Setting: {proj} Matches Board Setting: {board}.'
                                          .format(key=key,
                                                  proj=getattr(top_level_dependency, key, False),
                                                  board=self.board_settings[key]))
                        setattr(self, key, self.board_settings[key])

                    else:
                        self.logger.warning('[{key}] Project Setting: {proj} Does Not Matches Board Setting: '
                                            '{board}. Using Board Setting.'
                                            .format(key=key,
                                                    proj=getattr(top_level_dependency, key, False),
                                                    board=self.board_settings[key]))

                        setattr(self, key, self.board_settings[key])

            for key in self.board_settings:
                if not hasattr(self, key):
                    self.logger.info('[{key}] Adding Setting from: {board_settings_file}'
                                     .format(key=key, board_settings_file=board_settings_file))
                    setattr(self, key, self.board_settings[key])

            if hasattr(self, 'ip'):
                setattr(self, 'has_vendor_ip', True)
            else:
                setattr(self, 'has_vendor_ip', False)

            if not hasattr(self, 'board_id'):
                setattr(self, 'board_id', 'X"0000_0000"')
            if not hasattr(self, 'board_timestamp_enabled'):
                setattr(self, 'board_timestamp_enabled', False)


class VendorIpDependency(object):
    """Vendor IP is an Vendor IP object containing methods and attributes required for processing ``vendor_ip``

    Keyword Args:
        **repo_root (str): Absolute path to the system environment ``$REPO_ROOT``.
        **master_vendor_ip (obj): A project dependency object for the *master* ``vendor_ip`` dependency for
            the project.
        **vendor (str): FPGA Vendor. Used to determine ``vendor_ip`` locations.
            Valid values: ``'xilinx'`` and ``altera'``. Default value: ``'xilinx'``
        **tool_version (str): FPGA Tool Version. Used to determine ``vendor_ip`` locations.
        **family (str): FPGA Family. Used to determine ``vendor_ip`` locations.
        **device (str): FPGA Device. Used to determine ``vendor_ip`` locations.
        **ip_dict (dict):
        **settings_file (str): Absolute path to settings YAML file, used in logs.
    """
    def __init__(self, **kwargs):
        self.logger = log.config_logger(name=__name__, class_name=self.__class__.__name__)

        setattr(self, 'repo_root', funcs.get_kwarg('repo_root', kwargs, None))
        setattr(self, 'vendor', funcs.get_kwarg('vendor', kwargs, 'xilinx'))
        setattr(self, 'tool_version', funcs.get_kwarg('tool_version', kwargs, None))
        setattr(self, 'family', funcs.get_kwarg('family', kwargs, None))
        setattr(self, 'device', funcs.get_kwarg('device', kwargs, None))
        ip_dict = funcs.get_kwarg('ip_dict', kwargs, dict())
        setattr(self, 'settings_file', funcs.get_kwarg('settings_file', kwargs, ''))
        master_vendor_ip = funcs.get_kwarg('master_vendor_ip', kwargs, None)

        master_layout_list = list()

        if master_vendor_ip:

            master_layout_list = self.resolved_master_tags(paths=master_vendor_ip.layout_list,
                                                           vendor=self.vendor,
                                                           tool_version=self.tool_version,
                                                           family=self.family,
                                                           device=self.device)

            for p in master_layout_list:
                self.logger.info('(Unresolved) Master Vendor IP Path: {p}'.format(p=p))

            resolved_ip = dict()
            for cat, subcats in ip_dict.items():
                if isinstance(subcats, (dict, )):
                    for subcat, v in subcats.items():
                        if isinstance(v, (dict, )):
                            try:
                                if v['enabled']:
                                    resolved_ip[subcat] = {'ipcat': cat,
                                                           'ipsubcat': '',
                                                           'lib': v['library'],
                                                           'wrapper_file': v['wrapper_file']}

                                    self.logger.debug('No Subcategory!: {cat}|{name}'
                                                      .format(cat=cat,
                                                              name=subcat))
                            except KeyError:
                                for n, s in v.items():
                                    try:
                                        if s['enabled']:
                                            resolved_ip[n] = {'ipcat': cat,
                                                              'ipsubcat': subcat,
                                                              'lib': s['library'],
                                                              'wrapper_file': s['wrapper_file']}

                                            self.logger.debug('Subcategory!: {cat}|{subcat}|{name}'
                                                              .format(cat=cat,
                                                                      subcat=subcat,
                                                                      name=n))
                                    except KeyError:
                                        self.logger.critical('Vendor IP Category/Sub-Category too Deep: '
                                                             '{cat}|{subcat}|{name}. Skipping'
                                                             .format(cat=cat,
                                                                     subcat=subcat,
                                                                     name=n))
            for k, v in resolved_ip.items():
                paths = list()

                for p in master_layout_list:
                    resolved_path = p.replace('<NAME>', k)
                    resolved_path = resolved_path.replace('<IPCAT>', v['ipcat'])
                    resolved_path = resolved_path.replace('<IPSUBCAT>', v['ipsubcat'])
                    paths.append(os.path.abspath(resolved_path))

                for p in master_layout_list:
                    if 'common_wrapper_files' in p:
                        try:
                            wrapper_name = v['wrapper_file'].replace('_wrapper', '')
                            resolved_path = p.replace('<NAME>', wrapper_name)
                            resolved_path = resolved_path.replace('<IPCAT>', v['ipcat'])
                            resolved_path = resolved_path.replace('<IPSUBCAT>', v['ipsubcat'])
                            if os.path.abspath(resolved_path) not in paths:
                                paths.append(os.path.abspath(resolved_path))
                        except AttributeError:
                            pass

                self.logger.debug('All Found Paths for {ip}:'
                                  .format(ip=k))
                for path in paths:
                    self.logger.info('(Resolved) IP Search Path: {path}'
                                     .format(path=path))

                for path in paths:
                    for root, subpaths, filenames in os.walk(path):
                        self.logger.debug('Subpaths: {subpaths}'.format(subpaths=subpaths))
                        top_level_paths = [os.path.join(root, sp) for sp in subpaths]
                        for top in top_level_paths:
                            self.logger.debug('\tTop-Level Path: {top}'
                                              .format(top=top))
                            top_dir_str = ProjectDependency.top_dir_lookup(dir_name=os.path.split(top)[1])
                            if top_dir_str in resolved_ip[k]:
                                resolved_ip[k][top_dir_str].append(top)
                            else:
                                resolved_ip[k][top_dir_str] = [top]
                        break

            setattr(self, 'enabled_ip', resolved_ip)

    def resolved_master_tags(self, **kwargs):
        """

        Keyword Args:
            **paths (str, or list of str): Absolute path to resolve tags
            **vendor (str): FPGA Vendor. Used to determine ``vendor_ip`` locations.
                Valid values: ``'xilinx'`` and ``altera'``. Default value: ``'xilinx'``
            **tool_version (str): FPGA Tool Version. Used to determine ``vendor_ip`` locations.
            **family (str): FPGA Family. Used to determine ``vendor_ip`` locations.
            **device (str): FPGA Device. Used to determine ``vendor_ip`` locations.

        Returns:
            list of str: List of absolute paths with tags resolved.
        """

        paths = funcs.get_kwarg('paths', kwargs, None)
        vendor = funcs.get_kwarg('vendor', kwargs, 'xilinx')
        tool_version = funcs.get_kwarg('tool_version', kwargs, None)
        family = funcs.get_kwarg('family', kwargs, None)
        device = funcs.get_kwarg('device', kwargs, None)

        if isinstance(paths, (str, )) and paths != "":
            paths = [paths]

        resolved_paths = list()

        for p in paths:
            untagged_path = p.replace('<VENDOR>', vendor)
            untagged_path = untagged_path.replace('<TOOL_VERSION>', tool_version)
            untagged_path = untagged_path.replace('<FAMILY>', family)
            untagged_path = untagged_path.replace('<DEVICE>', device)
            split_root = os.path.join(untagged_path.split('<NAME>')[0], '<NAME>')
            if split_root not in resolved_paths:
                resolved_paths.append(split_root)

        return resolved_paths


if __name__ == '__main__':
    logger.info("projectmanager {} {}"
                .format(__version__, version.revision))
    log.sect_break(logger)

    project = RepoFlow()

else:
    logger.info(__str__)



# -*- coding: utf8 -*-
"""

************************
``helpers/arguments.py``
************************

`BSD`_ © 2018-2019 Science and Technology Facilities Council & contributors

.. _BSD: _static/LICENSE

``arguments.py`` is a helpers module provided to define and configure the argument passing and parsing for the
main modules.


"""
import os
import yaml
import argparse

import customlogging as log
logger = log.config_logger(name=__name__)

import version as my_version
rev = filter(str.isdigit, "$Rev$")  # Do Not Modify this Line
version = my_version.Version(0, 8, 0, svn_rev=rev, disable_svn_logging=True)
__version__ = version.get_version()
__str__ = my_version.about(__name__, __version__, version.revision)


class Arguments(object):
    """Creates arguments for passing files and configuration data to the script.

    A Class which configures and retrieves Arguments for use by a python script. Dictionaries
    retrieved from YAML files are added as attributes to this class with no sub-processing.
    Information including the filename of each YAML file parsed is also added as an attribute.

    See :ref:`arguments`

    Attributes:

        config (dict): the parsed dictionary of the ``config.yml`` passed using to script ``--config``
        settings (dict): the parsed dictionary of the ``settings.yml`` passed to script using ``--settings``
        checkout_enabled (bool): set ``True`` using optional ``--checkout_enabled``, otherwise ``False``
        dev_flag (bool): set ``True`` using optional ``--dev``, otherwise ``False``
        open_gui (str): Valid values: ``'fpga'``, ``'sim'``. Using optional ``--open_gui``, otherwise ``None``
        clean (list of str): Valid values: ``'all'``, ``'fpga'``, ``'ip'``, ``'precompiled_sim'``, ``'sim'``.
            Using optiona ``clean``, otherwise: ``None``

    """
    def __init__(self):
        self.logger = log.config_logger(name=__name__, class_name=self.__class__.__name__)
        _parser = self._create_parser()
        self._add_arguments(_parser)
        args = self._get_args(_parser)
        for k in vars(args):
            self.logger.info('Processed Argument: --{arg}'
                             .format(arg=k))

        log.sect_break(self.logger)
        self.logger.info('Loading Arguments from Command Line...')
        log.sect_break(self.logger)
        self._import_from_yaml(args)
        try:
            self._vunit_args(self.settings, self.settings_file)
        except AttributeError:
            self.logger.warning("Settings Missing. Using Default Options for VUnit.")

    def __repr__(self):
        return "{self.__class__.__name__}()".format(self=self)

    def __str__(self):
        return "{} is a {self.__class__.__name__} object".format(__name__, self=self)

    @staticmethod
    def _create_parser():
        """

        Creates an ``argparse`` object

        :return:
        """
        return argparse.ArgumentParser()

    @staticmethod
    def _add_arguments(parser):
        """Adds Arguments to provide to python script, using the argparse module.

        Args:
            parser (obj): ``argparse`` object

        Returns:
            None

        """
        parser.add_argument("--config",
                            dest="config",
                            type=argparse.FileType('r'),
                            default="./config.yml",
                            required=True,
                            help="YAML File Defining Project")
        parser.add_argument("--settings",
                            dest="settings",
                            type=argparse.FileType('r'),
                            required=True,
                            help="YAML File Defining Settings for Top-Level")
        parser.add_argument("--checkout_enabled",
                            dest="checkout_disabled",
                            action="store_false",
                            help="Enables Checkout When Creating RepoFlow Objects")
        parser.add_argument("--dev",
                            dest="dev_flag",
                            action="store_true",
                            help="Development Flag for Executing Script in PyCharm")
        parser.add_argument("--headless",
                            dest="headless",
                            action="store_true",
                            help="Headless Flag for Forcing GUI Based Options to False. Used to Ensure"
                                 "Continuous Integration Environments Execute Unimpeded by "
                                 "External User Input")
        parser.add_argument("--open_gui",
                            dest="open_gui",
                            choices=['fpga', 'sim'],
                            action="store",
                            default=None,
                            help="Opens Project in Specified Tool After Generating Project")
        parser.add_argument("--clean",
                            dest="clean",
                            choices=['all', 'fpga', 'ip', 'precompiled_sim', 'sim'],
                            nargs="*",
                            action="append",
                            default=None,
                            help="Cleans the Projects Specified Output Prior to Generating Project.")

    @staticmethod
    def _get_args(parser):
        """Fetches the arguments from a argparse object

        Args:
            parser (obj): argparse object

        Returns:
            None

        """
        args = parser.parse_args()
        return args

    def _import_from_yaml(self, args):
        """Imports attributes from ``yml`` file.

        Imports attributes from YAML Files defined by ``_ARG_FILELIST`` which have been passed to via
        Argument(s) attribute names match the corresponding arg name, and a reference containing the
        absolute path to the YAML file is also stored in an attribute named ``<arg>_file``

        Args:
            settings (dict):
            config (dict):

        returns:
            None

        """
        for arg in vars(args):
            attr = getattr(args, arg)
            if hasattr(attr, 'read'):
                name_attr = '{}'.format(arg)
                fname_attr = '{}_file'.format(arg)
                self.logger.info('Processing:                   --{}'
                                 .format(name_attr))
                self.logger.info('Filepath Reference Attribute: {}'
                                 .format(fname_attr))
                try:
                    abspath = os.path.abspath(attr.name)
                    contents = yaml.load(attr)
                    attr.close()
                    setattr(self, fname_attr, abspath)
                    self.logger.info('Path:                         {}'
                                     .format(abspath))
                    setattr(self, name_attr, contents)
                except AttributeError as e:
                    self.logger.error('Error Importing {} YAML File from Argument. File '
                                      '(Probably) Not a YAML File'
                                      .format(fname_attr))
                    self.logger.error('\t{}.'
                                      .format(e))
                    log.errorexit(self.logger)
            else:
                self.logger.info('Processing:                   --{}'
                                 .format(arg))
                setattr(self, arg, getattr(args, arg))

    def _vunit_args(self, settings, settings_file, vunit_args_id='vunit_args', headless=False):
        """Sets up VUnit Arguments to pass from settings.

        Replicates the passing of arguments from the command line.

        Args:
            settings (dict) Settings retrieved from YAML settings file
            settings_file (str): Absolute path of settings files used.
            vunit_args_id (str, optional): Key in settings dict representing VUnit arguments.
                Default value: ``'vunit_args'``

        Returns:
            None

        """
        supported_vunit_args = ['version', 'log_level',  'output_path', 'test_patterns',
                                'list', 'files', 'keep_compiling', 'num_threads',
                                'elaborate', 'verbose', 'xunit_xml', 'coverage',
                                'unique_sim', 'exit_0', 'no_color']

        if vunit_args_id in settings:
            self.logger.info("Appending VUnit Arguments using '{vunit_args}:' Key from Settings File: {file}"
                             .format(vunit_args=vunit_args_id,
                                     file=settings_file))
            for key, value in settings[vunit_args_id].items():
                if key in supported_vunit_args:
                    self.logger.debug("Setting VUnit Arg: {key}:{value}"
                                      .format(key=key, value=value))
                    setattr(self, key, value)
                    if key == 'version' and value:
                        self.logger.warning("Settings YAML File is Configuring VUnit to Only return Version")
                    elif key == 'list' and value:
                        self.logger.warning("Settings YAML File is Configuring VUnit to Only List Test-Cases")
                        self.logger.warning("\tOnly Test Cases Matching 'test_patterns' are Listed")
                    elif key == 'files' and value:
                        self.logger.warning("Settings YAML File is Configuring VUnit to Only List "
                                            "Compile Ordered Files")
                else:
                    if key == 'clean':
                        self.logger.critical('"vunit_args: {key}" is a Depreciated VUnit Option in Settings '
                                             'File: {settings}'
                                             .format(key=key, settings=settings_file))
                        self.logger.critical('\tInstead Use Argument: "--clean sim"')
                    elif key == 'clean_precompiled_libs':
                        self.logger.critical('"vunit_args: {key}" is a Depreciated VUnit Option in Settings '
                                             'File: {settings}'
                                             .format(key=key, settings=settings_file))
                        self.logger.critical('\tInstead Use Argument: "--clean precompiled_sim"')
                    elif key == 'gui':
                        self.logger.critical('"vunit_args: {key}" is a Depreciated VUnit Option in Settings '
                                             'File: {settings}'
                                             .format(key=key, settings=settings_file))
                        self.logger.critical('\tInstead Use Argument: "--open_gui sim"')
                    else:
                        self.logger.critical('"vunit_args: {key}: {value}" is an Unsupported VUnit Option in '
                                             'Settings File: {settings}'
                                             .format(key=key, value=value, settings=settings_file))
        else:
            self.logger.warning("VUnit Arguments '{vunit_args}' Missing from Settings. "
                                "Using Default Options"
                                .format(vunit_args=vunit_args_id))




# -*- coding: utf8 -*-
"""

**********************
``helpers/version.py``
**********************

`BSD`_ © 2018-2019 Science and Technology Facilities Council & contributors

.. _BSD: _static/LICENSE

``version.py`` is a helpers module provided to handle public version numbering which follows the
PEP440 guidelines.

.. note::

   The ``__version__`` uses the :mod:`version.Version` (defined in this module), therefore is used
   towards the end of this module.

"""
import customlogging as log
logger = log.config_logger(name=__name__)


def about(modname, ver, revision):
    """Method to Create an About String, to display Method Name and Version/Rev

    Args:
        modname (str): Module Name
        ver (str):  Version Number, in the form: ``MAJOR.MINOR.MICRO``
        revision (str): ``SVN`` Revision Number

    Returns:
        str:

    """
    return "{modname} {ver} {revision}".format(modname=modname, ver=ver, revision=revision)


class Version:
    """Creates a Class with methods to construct and fetch PEP440 Public Version Numbering

    Creates an initial PEP440 Public Version Identifier from the supplied parameters after validating the
    prerelease parameter,

    Args:
        major (int): Representing the Major Version Number (changes affecting execution method)
        minor (int): Representing the Minor Version Number (bugfixes)
        micro (int): Representing the Micro Version Number (correction of typos/release notes)
        prerelease (int, optional): Addition Pre-Release Indicator. Valid values: ``a``, ``b``, ``rc``,
            where: ``a`` = alpha; ``b`` = beta; and ``rc`` = release candidate
        svn_rev (str, optional): ``SVN`` Revision Number
        disable_svn_logging (bool): Do not include ``SVN`` output in logs. Default value: ``False``

    """
    def __init__(self, major, minor, micro, prerelease='', svn_rev='', disable_svn_logging=False):
        self.logger = log.config_logger(name=__name__, class_name=self.__class__.__name__)
        self._svn_rev = svn_rev
        self.revision = self._validate_svn_rev(self._svn_rev, disable_logging=disable_svn_logging)
        self._major = major
        self._minor = minor
        self._micro = micro
        self._pre = prerelease
        self.version = self.set_version(self._major, self._minor, self._micro, self._pre)

    def __repr__(self):
        return "{self.__class__.__name__}({self._major}, {self._minor}, " \
               "{self._micro}, prerelease='{self._pre}', svn_rev='{self._svn_rev}')".format(self=self)

    def __str__(self):
        return "{self.version} {self.revision}".format(self=self)

    def _validate_svn_rev(self, svn_rev, disable_logging=True):
        """
        Method which checks the SVN Revision Number (normally derived from parsing ``SVN`` ``REV`` Keyword
        Substitution and returns a string. Invalid and empty values return an empty str with a critical
        logger level message.

        If an empty str is passed to this method it usually means keyword substitution has not been enabled
        on the file calling this method, or the file calling this method has not been committed yet.

        :param svn_rev:
        :return: str
        """
        if svn_rev == '':
            if not disable_logging:
                self.logger.critical("Not a Valid SVN Revision No: {}"
                                     .format(svn_rev))
            rev_str = ''
        else:
            rev_str = "(r{})".format(svn_rev)

        return rev_str

    def _validate_prerelease(self, prerelease):
        """
        Method which checks the Pre-Release value is valid. Valid Pre-Release Indicators Are:
        * ``a`` = alpha
        * ``b`` = beta
        * ``rc`` = release candidate

        This checking is not case sensitive, but lower case values are returned. Invalid types and values
        return an empty str with a critical logger level message.

        :param prerelease:
        :return: str
        """
        try:
            if prerelease == '':
                self.logger.debug("Not a Pre-Release")
                pre = ''
            elif prerelease.lower() == 'a':
                self.logger.debug("An Alpha Release")
                pre = 'a'
            elif prerelease.lower() == 'b':
                self.logger.debug("A Beta Release")
                pre = 'b'
            elif prerelease.lower() == 'rc':
                self.logger.debug("A Release Candidate")
                pre = 'rc'
            else:
                self.logger.critical("Invalid prerelease Value: {}"
                                     .format(prerelease))
                pre = ''
        except AttributeError:
            self.logger.critical("prerelease Doesn't Appear to be a Str: {}"
                                 .format(prerelease))
            pre = ''

        return pre

    def set_version(self, major, minor, micro, pre):
        """

        Method to set the string representation of a Version Number.

        :param major: Integer Value Representing the Major Version Number (changes affecting execution method)
        :param minor: Integer Value Representing the Minor Version Number (bugfixes)
        :param micro: Integer Value Representing the Micro Version Number (correction of typos/release notes)

        :param pre: Optional Addition Pre-Release Indicator...
            a = alpha
            b = beta
            rc = release candidate

        :return: string
        """
        return "{}.{}.{}{}".format(major, minor, micro, self._validate_prerelease(pre))

    def get_version(self):
        """
        Method to get the string representation of the version number.

        :return: String of Version Number
        """
        return self.version


rev = filter(str.isdigit, "$Rev$")  # Do Not Modify this Line
version = Version(1, 1, 0, prerelease='rc', svn_rev=rev, disable_svn_logging=True)

__version__ = version.get_version()
__str__ = about(__name__, __version__, version.revision)



# -*- coding: utf8 -*-
"""

***********************
``fpgavendor_iface.py``
***********************

`BSD`_ © 2018-2019 Science and Technology Facilities Council & contributors

.. _BSD: _static/LICENSE

``fpgavendor_iface.py`` is a module to handle the generation of FPGA projects and interfacing with
:mod:`vunit_iface`, to include the compilation of FPGA vendor simulation modules based on tool version and
FPGA device family.

"""
import helpers.customlogging as log
import helpers.version as my_version
import helpers.funcs as funcs

import os
import shutil
import time
import datetime
from subprocess import check_call, CalledProcessError

logger = log.config_logger(name=__name__)
LOGLEVEL = log.LogLevelsConsts

rev = filter(str.isdigit, "$Rev$")  # Do Not Modify this Line
version = my_version.Version(0, 1, 0, svn_rev=rev, disable_svn_logging=True)
__version__ = version.get_version()
__str__ = my_version.about(__name__, __version__, version.revision)


class FpgaProject(object):
    """

    """
    def __init__(self, **kwargs):
        self.logger = log.config_logger(name=__name__, class_name=self.__class__.__name__)
        self.logger.info('*' * 80)
        self.logger.info('** Generating FPGA Project...')
        self.logger.info('*' * 80)
        setattr(self, 'repo_root', funcs.get_kwarg('repo_root', kwargs, ''))
        build_path = funcs.get_kwarg('build_path', kwargs, None)
        rebuild_project = funcs.get_kwarg('rebuild_project', kwargs, True)
        setattr(self, 'board_id', funcs.get_kwarg('board_id', kwargs, 0))
        setattr(self, 'board_timestamp_enabled', funcs.get_kwarg('board_timestamp_enabled', kwargs, False))
        setattr(self, 'tcl_repo_root', '$env(REPO_ROOT)')

        if not build_path:
            self.logger.error('No Build Path Supplied. Can Not Create {self.__class__.__name__}'
                              .format(self=self))
            log.errorexit(self.logger)

        project_name = funcs.get_kwarg('project_name', kwargs, 'project')

        generation_script_name = funcs.get_kwarg('generation_script_name', kwargs, 'project_gen.tcl')
        generation_script_ext = '.tcl'

        project_extension = os.path.splitext(project_name)[1]

        if not generation_script_name.endswith(generation_script_ext):
            generation_script_name = generation_script_name + generation_script_ext

        project_file = os.path.join(build_path, project_name)
        setattr(self, 'project_file', project_file)

        self.logger.info('Project Name:       {project_name}'
                         .format(project_name=project_name))
        self.logger.info('Project Build Path: {build_path}'
                         .format(build_path=build_path))

        family = funcs.get_kwarg('family', kwargs, None)
        device = funcs.get_kwarg('device', kwargs, None)

        top_level_entity = funcs.get_kwarg('top_level_entity', kwargs, '')
        if isinstance(top_level_entity, (list, )):
            top_level_entity = top_level_entity[0]

        self.logger.info('Top-Level Entity:   {top_level_entity}'
                         .format(top_level_entity=top_level_entity))

        hdl_src_files = funcs.get_kwarg('hdl_src_files', kwargs, list())
        ip_files = funcs.get_kwarg('ip_files', kwargs, list())
        constraint_files = funcs.get_kwarg('constraint_files', kwargs, list())

        if project_extension in ['.xpr']:
            vendor = 'xilinx'
            generate_script = self.create_vivado_project(project_name=project_name,
                                                         build_path=build_path,
                                                         top_level_entity=top_level_entity,
                                                         hdl_files=hdl_src_files,
                                                         ip_files=ip_files,
                                                         constraint_files=constraint_files,
                                                         family=family,
                                                         device=device)
        elif project_extension in ['.qpf', '.qsf']:
            vendor = 'altera'
            generate_script = self.create_quartus_project(project_name=project_name,
                                                          build_path=build_path,
                                                          top_level_entity=top_level_entity,
                                                          hdl_files=hdl_src_files,
                                                          ip_files=ip_files,
                                                          constraint_files=constraint_files,
                                                          family=family,
                                                          device=device)

        else:
            self.logger.error('Can Not Determine Vendor from Project File Extension: {file}'
                              .format(file=project_name))
            log.errorexit(self.logger)

        if generate_script:
            generation_script_file_path = build_path
            self.logger.info('Project Build Path: {path}'
                             .format(path=generation_script_file_path))

            if not os.path.exists(generation_script_file_path):
                self.logger.warning('Project Build Path Not Found. Creating: {path}'
                                    .format(path=generation_script_file_path))
                os.makedirs(generation_script_file_path)

            elif rebuild_project:
                self.logger.info('Backing Up Existing Project Build Path: {path}'
                                 .format(path=generation_script_file_path))
                self.backup_project(project_path=generation_script_file_path)

            else:
                self.logger.critical('Project Build Path Exists and Custom Changes Will be Overwritten '
                                     'Without Being Backed Up: {path}'
                                     .format(path=generation_script_file_path))
                self.logger.critical("\tTo Prevent This Warning Use Argument: '--clean fpga' "
                                     "or '--clean all'")
                funcs.clean_path(path=generation_script_file_path)

            self.logger.info('Creating Project Generation Script: {script}'
                             .format(script=generation_script_file_path))

            logger.info('Writing Project Generation Script File: {script}'
                        .format(script=generation_script_file_path))

            fp_generation_script = os.path.join(generation_script_file_path,
                                                generation_script_name)

            with open(fp_generation_script, 'w') as script_fp:
                for line in generate_script:
                    script_fp.write(line + '\n')

            if vendor == 'xilinx':
                run_vivado(vivado_bin='vivado',
                           mode='tcl',
                           tcl_file=generation_script_file_path)

            elif vendor == 'altera':
                saved_path = os.getcwd()
                os.chdir(generation_script_file_path)
                run_quartus_sh(tcl_file_name=fp_generation_script,
                               script_mode=True)
                os.chdir(saved_path)

        self.logger.info('** Finished Generating FPGA Project')
        self.logger.info('*' * 80)

    def get_project_file(self):
        """Gets the absolute path of the project file.

        Returns:
            str: Absolute path of project file. ``None`` if not found.

        """
        return getattr(self, 'project_file', None)

    def add_ip(self, **kwargs):
        """

        Keyword Args:
            **vunit_obj (obj): A :mod:`vunit_iface.VunitProject` object. Default value: ``self.vu``
            **project_file (str): Absolute path to project file.
            **rebuild_project (bool, optional): If ``True`` rebuild existing project.
                Default value: ``False``

        Returns:
            None

        """
        vunit_obj = funcs.get_kwarg('vunit_obj', kwargs, self.vu)
        project_file = funcs.get_kwarg('project_file', kwargs, False)
        rebuild_project = funcs.get_kwarg('rebuild_project', kwargs, False)

        if project_file and project_file != "":
            project_dir = os.path.split(project_file)[0]
            project_extension = os.path.splitext(project_file)[1]

            if os.path.exists(project_dir):
                self.logger.info('Found Project Directory: {project_dir}'
                                 .format(project_dir=project_dir))
                if rebuild_project:
                    self.backup_project_file(project_dir=project_dir)
                else:
                    self.logger.warning('Overwriting Project: {file}'
                                        .format(file=project_file))
            else:
                self.logger.info('Creating Project Directory: {project_dir}'
                                 .format(project_dir=project_dir))
                os.makedirs(project_dir)
        else:
            self.logger.error('Incorrect Project File Supplied: {file}'
                              .format(file=project_file))
            log.errorexit(self.logger)

        self.logger.info('Processing Project File for Simulation: {project_file}'
                         .format(project_file=project_file))

    def backup_project(self, **kwargs):
        """Backup project using timestamp.

        Appends time stamp to the project directory in the form: ``_%Y%m%d%H%M%S`` and uses
        ``shutil.copytree()`` to backup the project. Then cleans the original directory (using
        ``shutil.rmtree()`` and remaking the location using: ``os.makedirs()``)

        Keyword Args:
            **project_path (str): Absolute path to project root.

        Returns:
            None

        """
        project_path = funcs.get_kwarg('project_path', kwargs, False)
        self.logger.info('Backing Up Existing Project:')

        timestamp = time.time()
        backup_timestamp = datetime.datetime.fromtimestamp(timestamp).strftime('_%Y%m%d%H%M%S')
        backup_project_path = project_path + backup_timestamp
        self.logger.info('\tfrom: {project_path}'
                         .format(project_path=project_path))
        self.logger.info('\tto  : {project_path}'
                         .format(project_path=backup_project_path))

        shutil.copytree(project_path, backup_project_path)
        logger.info('Backed Up Project Directory and all its Contents to: {project_path}'
                    .format(project_path=backup_project_path))

        self.logger.info('Cleaning Existing Project: {project_path}'
                         .format(project_path=project_path))
        funcs.clean_path(path=project_path)

        self.logger.info('Cleaned Project: {project_path}'
                         .format(project_path=project_path))

    def create_vivado_project(self, **kwargs):
        """Creates a ``vivado`` project using supplied settings.


        Keyword Args:
            **project_name (str, optional): Name of project file to generate.
                Default value: ``'project'``
            **build_path (str): Absolute path to build location for generated project
            **target_language (str, optional): Target default language for the generated project.
                Default value: ``'VHDL'``
            **top_level_entity (str): Top-level HDL entity name.
            **hdl_files_dict (dict): Dictionary of HDL libraries (keys) and list of absolute HDL files.
            **ip_files (list of str): List of absolute paths fof IP files to add to project.
            **constraint_files (list of str): List of absolute paths of constraints to add to project.

        Returns:
            None

        """
        project_name, project_ext = os.path.splitext(funcs.get_kwarg('project_name', kwargs, 'project'))
        build_path = funcs.get_kwarg('build_path', kwargs, '').replace(self.repo_root, self.tcl_repo_root)
        family = funcs.get_kwarg('family', kwargs, None)
        device = funcs.get_kwarg('device', kwargs, None)
        target_language = funcs.get_kwarg('target_language', kwargs, 'VHDL')
        top_level_entity = funcs.get_kwarg('top_level_entity', kwargs, '')
        hdl_files = funcs.get_kwarg('hdl_files', kwargs, dict())
        ip_files = funcs.get_kwarg('ip_files', kwargs, list())
        constraint_files = funcs.get_kwarg('constraint_files', kwargs, list())

        self.logger.info('Generating Vivado Project...')
        self.logger.info('Project Name: {name}'.format(name=project_name))
        self.logger.info('FPGA Family: {family}'
                         .format(family=family))
        self.logger.info('FPGA Device: {device}'
                         .format(device=device))

        project_file = os.path.join(build_path, (project_name + project_ext))
        tcl_cmds_list = list(['#' * 80, '## Auto-Generated Script File, Do Not Modify', '#' * 80])
        tcl_cmds_list.append('create_project {project} -force -part {device}'
                             .format(project=project_file, device=device))
        tcl_cmds_list.append('set_property part {device} [current_project]'
                             .format(device=device))
        tcl_cmds_list.append('set_property target_language {target_language} [current_project]'
                             .format(target_language=target_language))
        tcl_cmds_list.append('set_property XPM_LIBRARIES {XPM_CDC XPM_FIFO XPM_MEMORY} [current_project]')

        tcl_cmds_list.append('#' * 80)
        tcl_cmds_list.append('## Adding Script Based Parameters:')
        tcl_cmds_list.append('#' * 80)

        tcl_cmds_list.append('set_property generic g_board_id={self.board_id} [current_project]'
                             .format(self=self))

        if self.board_timestamp_enabled:
            timestamp = generate_timestamp()
            self.logger.info('Generated Timestamp: {timestamp}'
                             .format(timestamp=generate_timestamp()))
            tcl_cmds_list.append('set_property generic g_board_timestamp={timestamp} [current_project]'
                                 .format(timestamp=timestamp))

        tcl_cmds_list.append('#' * 80)
        tcl_cmds_list.append('## Adding HDL Source File(s):')
        tcl_cmds_list.append('#' * 80)

        for src_file_obj in hdl_files:
            hdl_file = src_file_obj[1].replace(self.repo_root, self.tcl_repo_root)
            tcl_cmds_list.append('add_files -norecurse "{hdl_file}"'
                                 .format(hdl_file=hdl_file))
            tcl_cmds_list.append('set_property library {lib} [get_files {hdl_file}]'
                                 .format(lib=src_file_obj[0], hdl_file=hdl_file))

        tcl_cmds_list.append('#' * 80)
        tcl_cmds_list.append('## Adding IP Source File(s):')
        tcl_cmds_list.append('#' * 80)

        for ip in ip_files:
            ext = os.path.splitext(ip)[-1].lower()
            if ext == '.xci':
                tcl_cmds_list.append('add_files -norecurse "{ip}"'
                                     .format(ip=ip.replace(self.repo_root, self.tcl_repo_root)))
            else:
                self.logger.critical('Unsupported IP File Type ({ext}): {ip}'
                                     .format(ext=ext, ip=ip))

        tcl_cmds_list.append('#' * 80)
        tcl_cmds_list.append('## Adding Constraints File(s):')
        tcl_cmds_list.append('#' * 80)

        for constr in constraint_files:
            ext = os.path.splitext(constr)[-1].lower()
            if ext == '.xdc':
                tcl_cmds_list.append('add_files -fileset constrs_1 -norecurse "{constr}"'
                                     .format(constr=constr.replace(self.repo_root, self.tcl_repo_root)))
            else:
                self.logger.critical('Unsupported Constraint File Type ({ext}): {constr}'
                                     .format(ext=ext, constr=constr))

        # \todo Add Pre/Post Scripts Here.
        #   tcl_cmds_list.append('set_property STEPS.SYNTH_DESIGN.TCL.PRE "{pre_synth}" [get_runs synth_1]'
        #                        .format(pre_synth=pre_synth))
        #   tcl_cmds_list.append('set_property STEPS.SYNTH_DESIGN.TCL.POST "{post_synth}" [get_runs synth_1]'
        #                        .format(post_synth=post_synth))
        #   tcl_cmds_list.append('set_property STEPS.OPT_DESIGN.TCL.PRE "{pre_opt}" get_runs impl_1]'
        #                        .format(pre_opt=pre_opt))
        #   tcl_cmds_list.append('set_property STEPS.OPT_DESIGN.TCL.POST "{post_opt}" get_runs impl_1]'
        #                        .format(post_opt=post_opt))

        tcl_cmds_list.append('quit')

        self.logger.info('#' * 80)
        logger.info('Project Generation Script:')
        for line in tcl_cmds_list:
            self.logger.info(line)

        self.logger.info('#' * 80)

        return tcl_cmds_list

    def create_quartus_project(self, **kwargs):
        """Creates a ``quartus`` project using supplied settings.

        Copy constraints from ``.qsf`` file line-by-line directly, ignoring blank and commented lines,
        into project file replacing any instance of ``$REPO_PATH`` with ``$env(REPO_ROOT)``.
        Leaves the original constraints file untouched, which should have the same effect as
        having sourced the constraints file.

        Keyword Args:
            **project_name (str, optional): Name of project file to generate.
                Default value: ``'project'``
            **build_path (str): Absolute path to build location for generated project
            **target_language (str, optional): Target default language for the generated project.
                Default value: ``'VHDL_2008'``
            **top_level_entity (str): Top-level HDL entity name.
            **hdl_files_dict (dict): Dictionary of HDL libraries (keys) and list of absolute HDL files.
            **ip_files (list of str): List of absolute paths fof IP files to add to project.
            **constraint_files (list of str): List of absolute paths of constraints to add to project.

        Returns:
            None

        """

        project_name = os.path.splitext(funcs.get_kwarg('project_name', kwargs, 'project'))[0]
        build_path = funcs.get_kwarg('build_path', kwargs, '').replace(self.repo_root, self.tcl_repo_root)
        family = funcs.get_kwarg('family', kwargs, None)
        device = funcs.get_kwarg('device', kwargs, None)
        target_language = funcs.get_kwarg('target_language', kwargs, 'VHDL_2008')
        top_level_entity = funcs.get_kwarg('top_level_entity', kwargs, '')
        hdl_files = funcs.get_kwarg('hdl_files', kwargs, dict())
        ip_files = funcs.get_kwarg('ip_files', kwargs, list())
        constraint_files = funcs.get_kwarg('constraint_files', kwargs, list())

        self.logger.info('Generating Quartus Project...')
        self.logger.info('Project Name: {name}'.format(name=project_name))
        self.logger.info('FPGA Family: {family}'
                         .format(family=family))
        self.logger.info('FPGA Device: {device}'
                         .format(device=device))

        tcl_cmds_list = list(['#' * 80, '## Auto-Generated Project File, Do Not Modify', '#' * 80])
        tcl_cmds_list.append('package require ::quartus::project')
        tcl_cmds_list.append('project_new -family "{family}" -part {device} {project_name}'
                             .format(family=family,
                                     device=device,
                                     project_name=project_name))

        tcl_cmds_list.append('set_global_assignment -name PROJECT_OUTPUT_DIRECTORY {output_path}'
                             .format(output_path=os.path.join(build_path, 'output')))
        tcl_cmds_list.append('set_global_assignment -name VHDL_INPUT_VERSION {vhdl_version}'
                             .format(vhdl_version=target_language))

        tcl_cmds_list.append('set_global_assignment -name TOP_LEVEL_ENTITY {top_level}'
                             .format(top_level=top_level_entity.replace("'", "")))

        tcl_cmds_list.append('#' * 80)
        tcl_cmds_list.append('## Adding Script Based Parameters:')
        tcl_cmds_list.append('#' * 80)

        tcl_cmds_list.append('set_parameter -name g_board_id {self.board_id}'
                             .format(self=self))

        if self.board_timestamp_enabled:
            timestamp = generate_timestamp()
            self.logger.info('Generated Timestamp: {timestamp}'
                             .format(timestamp=generate_timestamp()))
            tcl_cmds_list.append('set_parameter -name g_board_timestamp {timestamp}'
                                 .format(timestamp=timestamp))

        tcl_cmds_list.append('#' * 80)
        tcl_cmds_list.append('## Adding HDL Source File(s):')
        tcl_cmds_list.append('#' * 80)
        for src_file_obj in hdl_files:
            tcl_cmds_list.append('set_global_assignment -name VHDL_FILE {hdl_file} -library {lib}'
                                 .format(hdl_file=src_file_obj[1].replace(self.repo_root, self.tcl_repo_root),
                                         lib=src_file_obj[0]))

        tcl_cmds_list.append('#' * 80)
        tcl_cmds_list.append('## Adding IP Source File(s):')
        tcl_cmds_list.append('#' * 80)

        for ip in ip_files:
            ext = os.path.splitext(ip)[-1].lower()
            if ext == '.qsys':
                tcl_cmds_list.append('set_global_assignment -name QSYS_FILE {ip}'
                                     .format(ip=ip.replace(self.repo_root, self.tcl_repo_root)))
            elif ext == '.ip':
                tcl_cmds_list.append('set_global_assignment -name IP_FILE {ip}'
                                     .format(ip=ip.replace(self.repo_root, self.tcl_repo_root)))
            elif ext == '.qip':
                tcl_cmds_list.append('set_global_assignment -name QIP_FILE {ip}'
                                     .format(ip=ip.replace(self.repo_root, self.tcl_repo_root)))
            else:
                self.logger.critical('Unsupported IP File Type ({ext}): {ip}'
                                     .format(ext=ext, ip=ip))

        tcl_cmds_list.append('#' * 80)
        tcl_cmds_list.append('## Adding Constraints File(s):')
        tcl_cmds_list.append('#' * 80)

        for constr in constraint_files:
            ext = os.path.splitext(constr)[-1].lower()
            if ext == '.qsf':
                with open(constr) as f:
                    for line in f:
                        if line.strip():
                            for r in [self.repo_root, '$REPO_ROOT']:
                                modified_line = line.rstrip().replace(r, self.tcl_repo_root)

                            if modified_line.startswith('##'):
                                self.logger.debug('{old} -> {new}'
                                                  .format(old=line, new=modified_line))
                                tcl_cmds_list.append(modified_line)
                            elif not modified_line.startswith('#'):
                                self.logger.debug('{old} -> {new}'
                                                  .format(old=line, new=modified_line))
                                tcl_cmds_list.append(modified_line)

            elif ext == '.sdc':
                tcl_cmds_list.append('set_global_assignment -name SDC_FILE {constr}'
                                     .format(constr=constr.replace(self.repo_root, self.tcl_repo_root)))
            elif ext == '.tcl':
                tcl_cmds_list.append('set_global_assignment -name TCL_SCRIPT_FILE {constr}'
                                     .format(constr=constr.replace(self.repo_root, self.tcl_repo_root)))
            else:
                self.logger.critical('Unsupported Constraint File Type ({ext}): {constr}'
                                     .format(ext=ext, constr=constr))

        # \todo Add Pre/Post Scripts Here.
        tcl_cmds_list.append('export_assignments')

        self.logger.info('#' * 80)
        logger.info('Project Generation Script:')
        for line in tcl_cmds_list:
            self.logger.info(line)

        self.logger.info('#' * 80)

        return tcl_cmds_list


def generate_timestamp():
    """Generates a Unix 32 Bit Time Stamp from the systems time as an integer value.

    Returns:
        str: Timestamp as an integer for conversion to a VHDL `unsigned` ``std_logic_vector(31 downto 0)``.

    """
    return int(time.time())


def resolve_tool_version(**kwargs):
    """Resolves the tool_version based on vendor

    Keyword Args:
        tool_version (str): Version of tool to resolve.
        vendor (str, optional): Vendor name. Valid values: ``'altera'``, ``'xilinx'``. Default ``'xilinx'``

    Returns:
        str: The resolved ``tool_version`` when successful, otherwise ``False``

    """
    vendor = funcs.get_kwarg('vendor', kwargs, 'xilinx')
    tool_version = funcs.get_kwarg('tool_version', kwargs, None)

    if tool_version:
        if vendor == 'altera':
            return 'quartus_{version}_{release}'.format(version=os.path.split(tool_version)[0],
                                                        release=os.path.split(tool_version)[1])
        else:
            return 'vivado_{release}'.format(release=tool_version)
    else:
        return False


def run_vivado(**kwargs):
    """Runs ``vivado`` in specified mode.

    When running in ``'gui'`` mode, the tcl_file keyword argument is the project file to open.

    .. note::

       The ``shell=True`` in ``check_call`` is important in windows where ``vivado`` is just a bat file.

    Keyword Args:
        **vivado_bin (str, optional): Name of executable to run vivado. Default value: ``'vivado'``
        **mode (str, optional): Name of the mode to run. Valid values: ``'batch'``, ``'tcl'`` or ``'gui'``.
            Default value: ``'batch'``
        **tcl_file_name (str): Absolute path to the ``tcl`` script used by ``vivado`` when running in
            ``'tcl'`` mode, and the project file to open when running in ``'gui'`` mode.
        **tcl_args (str, optional): Additional positional arguments needed to execute ``tcl_file_name``
            Default value: ``None``

    Returns:
        None

    """
    vivado_bin = funcs.get_kwarg('vivado_bin', kwargs, 'vivado')
    mode = funcs.get_kwarg('mode', kwargs, 'batch')
    tcl_file_name = funcs.get_kwarg('tcl_file_name', kwargs, None)
    tcl_args = funcs.get_kwarg('tcl_args', kwargs, None)

    cmd = "{vivado_bin} -nojournal -nolog -notrace".format(vivado_bin=vivado_bin)
    cmd += " -mode {mode} ".format(mode=mode)
    if mode == 'gui':
        cmd += " {tcl_file}".format(tcl_file=os.path.abspath(tcl_file_name))
    else:
        cmd += " -source {tcl_file}".format(tcl_file=os.path.abspath(tcl_file_name))

        if tcl_args is not None:
            cmd += " -tclargs " + " ".join([str(val) for val in tcl_args])

    logger.info('Executing: {cmd}'.format(cmd=cmd))
    check_call(cmd, shell=True)


def run_quartus_sh(**kwargs):
    """Runs ``quartus_sh`` to execute compilation for standard HDL simulation libraries.

    Keyword Args:
        **quartus_bin (str, optional): Name of executable to run quartus. Default value: ``'quartus_sh'``
        **script_mode (bool, optional): Run ``quartus_sh`` in script mode.
            Default value: ``False``
        **
        **tool (str): Name of the tool to run. Should be derived from:
            ``simulator_class = SIMULATOR_FACTORY.select_simulator()``
            ``simname = simulator_class.name``
        **language (str, optional): Valid options: ``'verilog'`` or ``'vhdl'``. Default value: ``'vhdl'``
        **modelsim_path (str): Absolute path to the version of Modelsim to use to compile standard HDL
            libraries.
        **output_path (str): Absolute path where to compile standard HDL libraries.
        **tcl_file_name (str): Absolute path to the ``tcl`` script used by ``vivado``
        **tcl_args (str, optional): Additional positional arguments needed to execute ``tcl_file_name``
            Default value: ``None``
        **gui (bool, optional): Runs Quartus in GUI Mode. Default value: ``False``
        **project_file (str): Absolute path to project file.

    Returns:
        None

    """
    quartus_bin = funcs.get_kwarg('quartus_bin', kwargs, "quartus_sh")
    script_mode = funcs.get_kwarg('script_mode', kwargs, False)
    gui = funcs.get_kwarg('gui', kwargs, False)
    tool = funcs.get_kwarg('tool', kwargs, "quartus")
    language = funcs.get_kwarg('language', kwargs, "vhdl")
    modelsim_path = funcs.get_kwarg('modelsim_path', kwargs, None)
    output_path = funcs.get_kwarg('output_path', kwargs, None)
    tcl_file_name = funcs.get_kwarg('tcl_file_name', kwargs, None)
    tcl_args = funcs.get_kwarg('tcl_args', kwargs, None)
    project_file = funcs.get_kwarg('project_file', kwargs, None)

    if gui:
        cmd = "{quartus_bin}".format(quartus_bin=quartus_bin)
        if project_file:
            cmd += " {project_file}".format(project_file=project_file)

    elif not script_mode:
        cmd = "{quartus_bin} --simlib_comp".format(quartus_bin=quartus_bin)
        cmd += " -tool {tool}".format(tool=tool)
        cmd += " -language {language}".format(language=language)
        cmd += " -tool_path {modelsim_path}".format(modelsim_path=modelsim_path)
        cmd += " -directory {output_path}".format(output_path=output_path)
        cmd += " -family all"
        cmd += " -rtl_only"
    else:
        logger.info('Running {quartus_bin} in Script Mode...'
                    .format(quartus_bin=quartus_bin))
        cmd = "{quartus_bin} -t {tcl_file_name} -v".format(quartus_bin=quartus_bin,
                                                           tcl_file_name=tcl_file_name)

        if tcl_args is not None:
            cmd += " -tclargs " + " ".join([str(val) for val in tcl_args])

    logger.info('Executing: {cmd}'.format(cmd=cmd))
    check_call(cmd, shell=True)


def run_ip_setup_simulation(**kwargs):
    """Executes the Quartus ``run-ip-setup-simulation``

    Executes the ``run-ip-setup-simulation`` to gather simulation setup commands for all Vendor IP libraries
    used in current design.

    Keyword Args:
        **exe (str, optional): Executable to run. Default value: ``'ip-setup-simulation'``
        **project_file (str): Absolute path to project file to process.
        **output_path (str): Absolute path where output files are generated.

    Returns:
        None

    """

    exe = funcs.get_kwarg('exe', kwargs, "ip-setup-simulation")
    project_file = funcs.get_kwarg('project_file', kwargs, None)
    output_path = funcs.get_kwarg('output_path', kwargs, None)

    if not project_file:
        logger.critical('No Project File Supplied. Vendor IP Simulation Files Not Processed')
        return None

    if not output_path:
        logger.critical('No Output Path Supplied. Vendor IP Simulation Files Not Processed')
        return None

    if os.path.isfile(project_file):
        cmd = "{exe}".format(exe=exe)
        cmd += " --quartus-project={project_file}".format(project_file=project_file)
        cmd += " --output-directory={output_path}".format(output_path=output_path)

        logger.info('Executing: {cmd}'.format(cmd=cmd))
        try:
            check_call(cmd, shell=True)
        except CalledProcessError:
            logger.critical('Failed to Generate Simulation Files for Vendor IP (see Report Above): '
                            '{project_file}'
                            .format(project_file=project_file))
    else:
        logger.critical('Can Not Find Project File: {project_file}'
                        .format(project_file=project_file))


def run_ip_generate(**kwargs):
    """Executes the Quartus ``qsys-generate``

    Executes the ``qsys-generate`` to generate Quartus IP.

    Keyword Args:
        **exe (str, optional): Executable to run. Default value: ``'qsys-generate'``
        **ip_file (str): Absolute path to IP file to generate.
        **output_path (str): Absolute path where output files are generated.
        **synth (bool, optional): generate for synthesis if ``True`` otherwise generate for simulation.
            Default value: ``True``
        **synth_language (str, optional): HDL language used to generate synthesis files.
            Valid values: ``'VHDL'`` or ``'VERILOG'``. Default value: ``'VERILOG'``
        **sim_language (str, optional): HDL language used to generate simulation files.
            Valid values: ``'VHDL'`` or ``'VERILOG'``. Default value: ``'VHDL'``
        **family (str, optional): FPGA Family. Default value: ``False``
        **device (str, optional): FPGA Device. Default value: ``False``
        **clean (bool, optional): Cleans IP Path Prior to Generating IP. Default value: ``False``

    Returns:
        None

    """
    exe = funcs.get_kwarg('exe', kwargs, "qsys-generate")
    ip_file = funcs.get_kwarg('ip_file', kwargs, None)
    synth = funcs.get_kwarg('synth', kwargs, True)
    synth_language = funcs.get_kwarg('synth_language', kwargs, 'VERILOG')
    sim_language = funcs.get_kwarg('sim_language', kwargs, 'VHDL')
    family = funcs.get_kwarg('family', kwargs, False)
    device = funcs.get_kwarg('device', kwargs, False)
    output_path = funcs.get_kwarg('output_path', kwargs, None)
    clean = funcs.get_kwarg('clean', kwargs, False)

    if not ip_file:
        logger.critical('No IP File Supplied. Vendor IP Simulation Files Not Generated')
        return None

    if not output_path:
        logger.critical('No Output Path Supplied. Vendor IP Not Generated')
        return None

    if clean:
        logger.info('Cleaning IP Output Path: {output_path}'.format(output_path=output_path))
        funcs.clean_path(path=output_path)

    if os.path.isfile(ip_file):
        cmd = "{exe}".format(exe=exe)
        cmd += " {ip_file}".format(ip_file=ip_file)
        if synth:
            cmd += " --synthesis={synth_language}".format(synth_language=synth_language)
        else:
            cmd += " --simulation={sim_language}".format(sim_language=sim_language)
            cmd += " --allow-mixed-language-simulation"
        cmd += " --output-directory={output_path}".format(output_path=output_path)
        if family:
            cmd += ' --family="{family}"'.format(family=family)
        if device:
            cmd += ' --part="{device}"'.format(device=device)

        logger.info('Executing: {cmd}'.format(cmd=cmd))
        try:
            check_call(cmd, shell=True)
        except CalledProcessError:
            logger.critical('Vendor IP Not Generated (see Report Above): {ip_file}'
                            .format(ip_file=ip_file))
    else:
        logger.critical('Can Not Find Vendor IP File: {ip_file}'
                        .format(ip_file=ip_file))


if __name__ == '__main__':
    pass
else:
    logger.info(__str__)

